#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy import pi
import matplotlib.pyplot as plt
import warnings
from scipy.special import erfc
    

class PhotonFieldSolverOptimalWaveForm:
    
    def __init__(self, qubit, filterDrive=False):
        '''
        This calculates:
        inter-readout-cavity field for 0 and 1 state:
        self.resp_alpha0, self.resp_alpha1
        inter-filter-cavity field for 0 and 1 state:
        self.resp_beta0, self.resp_beta1
        
        
        Arguments:
            qubit :  a PurcellFilterQubit instance
            filterDrive :  Should the readout be driven through the readout line? (boolean)
        
        '''
        self.qubit = qubit
        self.params = qubit.params
        self.times = qubit.times
        
        #Change parameters for directly or indirectly driven readout
        if filterDrive:
            vfODE = vectorfields_filterDrive
        else:
            raise NotImplementedError("Resonator drive is not implemented yet!")
            
        self.edrive, self.resp_alpha0, self.resp_beta0, self.resp_alpha1, self.resp_beta1, self.c = solve_ODE_peak_and_hold(qubit=qubit, func=vfODE)
    

    def get_intercavity_fields(self, er=None, ef=None):
        er, ef =  self.get_amplitudes(ef=ef, er=er)
        return er*self.resp_alpha0, er*self.resp_alpha1
    
    def get_amplitudes(self, ef=None, er=None):
        if er is None:
            er = ef*(-1j*self.params.t_factor*self.params.G)
        elif ef is None:
            ef = er/(-1j*self.params.t_factor*self.params.G)
        return er, ef
    
    def get_filter_fields(self,  er=None, ef=None):
        er, ef = self.get_amplitudes(er=er, ef=ef)
        return np.array(er*self.resp_beta0), np.array(er*self.resp_beta1)
    
    def plot_alphas_time(self, er=None, ef=None, plot_imag=True):
        r0, r1 = self.get_intercavity_fields(er=er, ef=ef)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\alpha_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\alpha_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\alpha_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\alpha_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()
    
    def plot_betas_time(self, er=None, ef=None, plot_imag=True):
        r0, r1 = self.get_filter_fields(er=er, ef=ef)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\beta_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\beta_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\beta_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\beta_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()

    def plot_alphas_complex(self, er=None, ef=None):
        a0, a1 = self.get_intercavity_fields(er=er, ef=ef)

        x0 = np.real(a0)
        y0 = np.imag(a0)
        x1 = np.real(a1)
        y1 = np.imag(a1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.1
        xd = max(abs(xmin), abs(xmax))*0.1
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\alpha_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\alpha_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()


    def plot_betas_complex(self, er=None, ef=None):
        b0, b1 = self.get_filter_fields(er, ef)
        
        x0 = np.real(b0)
        y0 = np.imag(b0)
        x1 = np.real(b1)
        y1 = np.imag(b1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.4
        xd = max(abs(xmin), abs(xmax))*0.4
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\beta_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\beta_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()
    

    def plot_optimal_weight_function(self, e0=1, iq=False):
        r0, r1 = self.get_intercavity_fields(er=e0)

        weight_function = r1 - r0
        self.weight_function_I = weight_function.real
        self.weight_function_Q = weight_function.imag
            
        if iq:
            plt.plot(weight_function.real, weight_function.imag)
            plt.xlabel(r"Re($\alpha_e - \alpha_g$)")
            plt.ylabel(r"Im($\alpha_e - \alpha_g$)")
        else:
            plt.plot(self.times, weight_function.real, label=r"Re($\alpha_e - \alpha_g$)")
            plt.plot(self.times, weight_function.imag, label=r"Im($\alpha_e - \alpha_g$)")
            plt.xlabel("Time [ns]")
            plt.legend()
    
    def get_SNR_time(self, er=None, ef=None, eta=1, weightFct='optimal'):
        '''
        Set er OR ef for resonator drive or filter drive amplitude
        eta = Quantum efficieny (<=1)
        weightFct =     One of the following:
                        -'optimal' for the optimal weight function,
                        -a function working on the time array, returning
                        two arrarys describing the weight functions for I and Q,
                        -a Number being an even weight for I, Q and all times
                        -a two-touple being a weight for I and Q repectively.
        '''
        f0, f1 = self.get_filter_fields(er=er, ef=ef)
        
        a = 2 * np.sqrt(self.params.kappa_r * eta)
        diff = (f0 - f1)
        v_I = np.abs(np.real(diff))
        v_Q = np.abs(np.imag(diff))
        
        if weightFct == 'optimal':
            weight_I = v_I
            weight_Q = v_Q
            #print('Optimal')
        elif callable(weightFct):
            weight_I, weight_Q = weightFct(self.times)
            #print('User-Fct')
        elif isArray(weightFct) and len(weightFct) == 2:
            weight_I = np.ones_like(diff, dtype=float) * weightFct[0]
            weight_Q = np.ones_like(diff, dtype=float) * weightFct[1]
            #print('Two factors')
        else:
            weight_I = np.ones_like(diff, dtype=float) * weightFct
            weight_Q = np.ones_like(diff, dtype=float) * weightFct
            #print('Single factor')
            
        d_sig_t = a*((v_I*weight_I) + (v_Q*weight_Q)) * self.params.dt
        d_noise_t = ((weight_I**2) + (weight_Q**2)) * self.params.dt
        
        sig_t = np.cumsum(d_sig_t)
        noise_t_squared = np.cumsum(d_noise_t)
        noise_t = np.sqrt(noise_t_squared)
        
        SNR = np.zeros_like(self.times, dtype=float)
        d = np.where(noise_t != 0)
        SNR[d] = (sig_t[d] / noise_t[d])
        
        dSNR = np.zeros_like(self.times, dtype=float)
        dSNR[1:] = (SNR[1:] - SNR[:-1])/self.params.dt
        
        #SNR = np.cumsum(dSNR*self.params.dt) #Just for checkin'
        return (dSNR, SNR)
    
    def plot_SNR_time(self,er=None, ef=None, eta=1, naiive=False):
        dSNR, SNR = self.get_SNR_time(er=er, ef=ef, eta=eta)
        
        #fig=plt.figure(figsize=(9,4))
        fig, ax1 = plt.subplots(figsize=(9,4))
        ax1.plot(self.times*1e9, SNR, "-", color='red', label="Optimal SNR")
        ax1.set_xlabel('Time [ns]')
        ax1.set_ylabel('SNR')
        
        ax2 = plt.twinx()
        ax2.plot(self.times*1e9, dSNR, "--", color='red', label=r"Optimal $\frac{d SNR}{dt}$")
        ax2.set_ylabel(ylabel=r"$\frac{d SNR}{dt}$")
        
        if naiive:    
            dSNRn, SNRn = self.get_SNR_time(er=er, ef=ef, eta=eta, weightFct=(1,1))
            ax1.plot(self.times*1e9, SNRn, "-", color='blue', label="Naïve SNR")
            ax2.plot(self.times*1e9, dSNRn, "--", color='blue', label=r"Naïve  $\frac{d SNR}{dt}$")
        
        ax1.legend()
        ax2.legend()
        #plt.legend()
        
    def get_fidelity_time(self, er=None, ef=None, eta=1, weightFct='optimal'):
        dSNR, SNR = self.get_SNR_time(er=er, ef=ef, eta=eta, weightFct=weightFct)
        return (1-(erfc(SNR/2)/2))
        
    def plot_fidelity_time(self,er=None, ef=None, eta=1, naiive=False):
        fid = self.get_fidelity_time(er=er, ef=ef, eta=eta)
        #fig, ax1 = plt.subplots()
        plt.plot(self.times*1e9, fid, "-", color='red', label="Optimal SNR")
        plt.xlabel('Time [ns]')
        plt.ylabel('Fidelity')
        
        if naiive:    
            fidn = self.get_fidelity_time(er=er, ef=ef, eta=eta, weightFct=(1,1))
            plt.plot(self.times*1e9, fidn, "-", color='blue', label="Naïve SNR")
            
        plt.legend()
            
    def optimized_snr_slope(self, eta=1.0):
        "Obtain slope a of SNR using optimized weight functions"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(er=1)
        return 4*self.params.kappa_r*eta*np.sum(np.absolute(r0 - r1)**2)*self.params.dt
    
    def naive_weights_snr(self, eta=1.0, integration_factor=1):
        "obtain slope a of SNR using constant weight function (with optimal phase):"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(er=1)
        sig = 2*np.sqrt(eta*self.params.kappa_r) * np.absolute(np.sum((r0 - r1)))*self.params.dt
        noise = np.sqrt(self.params.tmax * integration_factor)
        return (sig/noise)**2
        
    def dephasing_strength(self):
        "Obtain dephasing rate b as"
        "contrast = exp(-b * e0**2)"        
        r0, r1 = self.get_intercavity_fields()
        return 2*self.chi*np.sum(np.imag(r1 * r0.conj())*self.params.dt)
    
    def estimate_eta(self, real_eta=1.0,  weight_method='optimal', integration_factor=1):
        if weight_method == 'optimal':
            snr_a = self.optimized_snr_slope(real_eta)
        else:
            snr_a = self.naive_weights_snr(real_eta, integration_factor)
        deph = self.dephasing_strength()
        return snr_a/(8*deph)
    

#####################################################################################
def avPhNum(f):
    n = np.average(np.abs(f)**2)
    return n

def isArray(a):
    return hasattr(a, '__len__')
#####################################################################################


def vectorfields_filterDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func]
    """
    alpha, beta = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(ef_func):
        ef = ef_func(t)
    elif isArray(ef_func):    
        l=len(ef_func)
        if  (l*1.01) > i:
            ef = ef_func[min(i,l-1)]
        else:
            ef = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        ef = ef_func
        
    #print(i, ef)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta
    db = - 1j*(w_f-w_d)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta - 1j*ef
    
    return [da, db]


def vectorfields_resonatorDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta, i]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func]
    """
    alpha, beta  = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(er_func):
        er = er_func(t)
    elif isArray(er_func):
        l=len(er_func)
        if  (l*1.01) > i:
            er = er_func[min(i,l-1)]
        else:
            er = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        er = er_func
    
    #print(i, er)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta - 1j*er
    db = - 1j*(w_f-w_d)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta
    
    return [da, db]


def solve_ODE_beat_up_and_down(qubit, func):
    drive = np.zeros_like(qubit.times, dtype=complex)
    resp0 = np.zeros_like([qubit.times]*2, dtype=complex)
    resp1 = np.zeros_like(resp0, dtype=complex)
    change = np.zeros_like(drive, dtype=complex)
    t=0
    
    #Constant parameters of the system
    dt = qubit.params.dt
    par0 = [qubit.params.w_d, qubit.params.w_r,
               qubit.params.w_f, qubit.params.chi,
               qubit.params.G, qubit.params.kappa_f,
               qubit.params.kappa_r, qubit.params.dt, 0j]
    par1 = par0[:] #Slice to copy!
    par1[3] *= -1 #Invert chi
    
    
    dri=0
    rampTime = 2.48e-9
    
    kr = 1/qubit.params.kappa_r
    kf = 1/qubit.params.kappa_f
    g = 1/qubit.params.g
    G = 1/qubit.params.G
    #kappa = max(kr, kf, g, G)
    kappa=(kr)
    
    f0 = tuple((0j,0j))
    f1 = tuple((0j,0j))
    d0 = tuple((0j,0j))
    d1 = tuple((0j,0j))
    maxPh = 0
    maxPhFilter = 0
    newP = 0
    newA = 0
    maxi = True
    w = 0j
    up = False
    
    
    for i, t in enumerate(qubit.times):
        photons = np.abs(f0[0])**2
        photonsFilter = np.abs(f0[1])**2
        maxPhFilter = max(photonsFilter, maxPhFilter)
        if maxi:
            maxPh = max(photons, maxPh)
        amp = 0.5*np.sqrt(photons)/kappa
        change[i] = amp
        
        #if kf > kr:
        if t < 2*rampTime:
            dri = np.exp(-1/2*((t-rampTime)/(rampTime/4))**2)
        else:
            dri = 0
            r=photons/maxPh
            if maxi:
                if r < 0.95:
                    maxi = False
            else:
                w = (f0[0]/abs(f0[0]))
                
                if t < newP+2*rampTime:
                    dri = newA*np.exp(-1/2*((t-newP)/(rampTime/4))**2)
                
                if (photonsFilter/maxPhFilter) < 0.001:
                    if t < newP+2*rampTime:
                        _=0
                    elif (up and (r < 0.999)):
                        direction = +1
                    elif (not(up) and (r > 0.2)):
                        direction = -1
                    elif (not(up) and (r < 0.1)):
                        direction = +1
                    elif (up and (r > 0.9)):
                        direction = -1
                
                    if direction == -1:
                        direction = 0
                        up = True
                        print("depl %4.f ns, %.3f"%(t*1e9, (photons/maxPh)))
                        newP = t + rampTime
                        A = min(np.sqrt(photons/maxPh), 1)
                        newA = np.conj(w)*A
                    elif direction == +1:
                        direction = 0
                        print("accu %4.f ns, %3.f"%(t*1e9, (photons/maxPh)))
                        up = False
                        newP = t + rampTime
                        A = np.sqrt((maxPh-photons)/maxPh)
                        newA = w*A
        #else:
        #    if t < rampTime*1.1:
        #        dri = np.exp(-5*((t-rampTime)/rampTime)**2)
        #    else:
        #        amp = 0.55*np.sqrt(photons)/kappa
        #        dri -= (dri-amp)*(dt/rampTime)
        #        #dri = maxAmp*np.cos(2*pi*(t-rampTime*1.1)/(kr+kf))**2
        #        #dri = maxAmp
        
        par0[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        par1[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        drive[i] = dri
        
        d0 = func(t=t, y=f0, par=par0)
        d1 = func(t=t, y=f1, par=par1)
        
        resp0[0,i] = f0[0]
        resp0[1,i] = f0[1]
        resp1[0,i] = f1[0]
        resp1[1,i] = f1[1]
        f0 = (f0[0] + d0[0]*dt, f0[1] + d0[1]*dt)
        f1 = (f1[0] + d1[0]*dt, f1[1] + d1[1]*dt)
    
    return drive, resp0[0,:], resp0[1,:], resp1[0,:], resp1[1,:], change



def solve_ODE_peak_and_hold(qubit, func):
    drive = np.zeros_like(qubit.times, dtype=complex)
    resp0 = np.zeros_like([qubit.times]*2, dtype=complex)
    resp1 = np.zeros_like(resp0, dtype=complex)
    change = np.zeros_like(drive, dtype=complex)
    t=0
    
    #Constant parameters of the system
    dt = qubit.params.dt
    par0 = [qubit.params.w_d, qubit.params.w_r,
               qubit.params.w_f, qubit.params.chi,
               qubit.params.G, qubit.params.kappa_f,
               qubit.params.kappa_r, qubit.params.dt, 0j]
    par1 = par0[:] #Slice to copy!
    par1[3] *= -1 #Invert chi
    
    
    dri=0
    rampTime = 2.48e-9
    
    kr = 1/qubit.params.kappa_r
    kf = 1/qubit.params.kappa_f
    g = 1/qubit.params.g
    G = 1/qubit.params.G
    #kappa = max(kr, kf, g, G)
    kappa=(kr)
    
    f0 = tuple((0j,0j))
    f1 = tuple((0j,0j))
    d0 = tuple((0j,0j))
    d1 = tuple((0j,0j))
    maxPh = 0
    
    
    for i, t in enumerate(qubit.times):
        photons = np.abs(f0[0])**2
        photonsFilter = np.abs(f0[1])**2
        amp = 0.9*np.sqrt(photons)/kappa
        change[i] = amp
        
        if t < 1.1*rampTime:
            dri = np.exp(-1/2*((t-rampTime)/(rampTime/4))**2)
        else:
            dri -= 2*(dri-amp)*(dt/rampTime)
        #else:
        #    if t < rampTime*1.1:
        #        dri = np.exp(-5*((t-rampTime)/rampTime)**2)
        #    else:
        #        amp = 0.55*np.sqrt(photons)/kappa
        #        dri -= (dri-amp)*(dt/rampTime)
        #        #dri = maxAmp*np.cos(2*pi*(t-rampTime*1.1)/(kr+kf))**2
        #        #dri = maxAmp
        
        par0[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        par1[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        drive[i] = dri
        
        d0 = func(t=t, y=f0, par=par0)
        d1 = func(t=t, y=f1, par=par1)
        
        resp0[0,i] = f0[0]
        resp0[1,i] = f0[1]
        resp1[0,i] = f1[0]
        resp1[1,i] = f1[1]
        f0 = (f0[0] + d0[0]*dt, f0[1] + d0[1]*dt)
        f1 = (f1[0] + d1[0]*dt, f1[1] + d1[1]*dt)
    
    return drive, resp0[0,:], resp0[1,:], resp1[0,:], resp1[1,:], change
