#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import warnings
from scipy.special import erfc, erf

class PurcellFilterQubit:
    '''
    Class for calculating the Green's functions and creatig the drive pulses.
    '''
    
    
    '''
    Array of floats (times), where pulses start and end
    '''
    significantTimes = None
    
    
    '''
    Array keeping the numerical values of the time windows (in secs).
    '''
    times = None
    
    '''
    Array keeping arrays of (dimensionless) drive pulse strengths of length
    len(times).
    '''
    windows = None
    
    '''
    Array of length len(times) keeping the numerical values of the Green's function
    for the readout-resonator for Qubit state 0.
    '''
    greens_function_alpha0 = None
    
    '''
    Array of length len(times) keeping the numerical values of the Green's function
    for the readout-resonator for Qubit state 1.
    '''
    greens_function_alpha1 = None
    
    
    '''
    Array of length len(times) keeping the numerical values of the Green's function
    for the filter for Qubit state 0.
    '''
    greens_function_beta0 = None
    
    
    '''
    Array of length len(times) keeping the numerical values of the Green's function
    for the filter for Qubit state 1.
    '''
    greens_function_beta1 = None
    
    '''
    Boolean: shall we use the simplified (fast-resonator-limit) or full interaction
    calculations?
    '''
    fullSim = False
    
    def __init__(self, params, fullSim=False, filterDrive=False):
        '''
            filterDrive :   Should the readout be driven through the readout
                            (or resonator) line? (boolean)
            fullSim :       Should we do a full or a simpler simulation,
                            where we assume the filter to be much faster than
                            the readout-resonator (boolean)
        '''
        params.is_well_defined() #Check if the parameters make sense.
        
        
        #Save some of the passed parameters
        self.fullSim = fullSim
        self.filterDrive = filterDrive
        self.params = params
        
        #Create the time-domain arrays
        self.times = np.arange(0, self.params.tmax, self.params.dt)
        self.ntimes = len(self.times)
        
        #
        self.windows = []
        self.windows.append(np.zeros(self.ntimes))
        
        
        self.significantTimes=[]
        
        if fullSim:
            '''
            The calculations for the full Green's functions are in my
            MEP diary on page 51.
            '''
            self.greens_function_alpha0, self.greens_function_alpha1, self.greens_function_beta0, self.greens_function_beta1 = self._full_greensFunctions()
            
        else:
            '''
            Simplified (fast-resonator-limit) Green's functions. Procedures
            are described in my MEP diary on page 51.
            '''
            self.greens_function_alpha0, self.greens_function_alpha1, self.greens_function_beta0, self.greens_function_beta1 = self._simple_greensFunctions()
        
    def _full_greensFunctions(self, times=None):
        if times is None:
            times = self.times
            
            
        G = self.params.G
        Gstar = self.params.Gstar
        
        wa = np.array([self.params.delta_rd +  self.params.chi, self.params.delta_rd - self.params.chi])
        wb = self.params.delta_fd - 1j*self.params.kappa_f/2
        
        s = np.sqrt((wa-wb)**2 + 4*np.abs(G)**2)
        wp = (wa + wb + s)/2
        wm = (wa + wb - s)/2
        
        
        I0 = 1j*(np.exp(-1j*wp[0]*times)-np.exp(-1j*wm[0]*times))/s[0]
        I1 = 1j*(np.exp(-1j*wp[1]*times)-np.exp(-1j*wm[1]*times))/s[1]
        J0 = 1j*(wp[0]*np.exp(-1j*wp[0]*times)-wm[0]*np.exp(-1j*wm[0]*times))/s[0]
        J1 = 1j*(wp[1]*np.exp(-1j*wp[1]*times)-wm[1]*np.exp(-1j*wm[1]*times))/s[1]
        
        if self.filterDrive:
            #Green's function of the resonator
            greens_function_alpha0 = 1j*G*I0
            greens_function_alpha1 = 1j*G*I1
            #Green's function of the filter
            greens_function_beta0 = 1j*J0 - 1j*wa[0]*I0
            greens_function_beta1 = 1j*J1 - 1j*wa[1]*I1
        else:
            #Green's function of the resonator
            greens_function_alpha0 = 1j*J0 - 1j*wb*I0
            greens_function_alpha1 = 1j*J1 - 1j*wb*I1
            #Green's function of the filter
            greens_function_beta0 = 1j*Gstar*I0
            greens_function_beta1 = 1j*Gstar*I1
            
        return greens_function_alpha0, greens_function_alpha1, greens_function_beta0, greens_function_beta1
            
    def _simple_greensFunctions(self, times=None):
        if times is None:
            times = self.times
        
        #Green's function of the resonator
        greens_function_alpha0 = -1j*np.exp((-1j*(self.params.delta_rd + self.params.delta_wr + self.params.chi) - self.params.kappa_eff/2)*times)
        greens_function_alpha1 = -1j*np.exp((-1j*(self.params.delta_rd + self.params.delta_wr - self.params.chi) - self.params.kappa_eff/2)*times)
        #Green's function of the filter
        greens_function_beta0 = -1j*np.exp((-1j*(self.params.delta_fd) - self.params.kappa_f/2)*times)
        greens_function_beta1 = -1j*np.exp((-1j*(self.params.delta_fd) -self. params.kappa_f/2)*times)
        
        return greens_function_alpha0, greens_function_alpha1, greens_function_beta0, greens_function_beta1
    
    def add_pulse(self, ts, te, f):
        '''
        Adds a drive pulse to the signal. All drive pulses will be added to
        make up the final drive signal.
        
        ts :    float; start time of the pulse in secs
        te :    float; end time of the pulse in secs
        f :     float; uniform amplitude in the specified time window OR
                callable; accepting an array of times and returning the according
                    amplitudes
        '''
        newWindow = self._new_window(ts=ts, te=te, f=f)
        self.windows.append(newWindow)
        
    
    def _new_window(self, ts, te, f):
        self.significantTimes.append(ts)
        self.significantTimes.append(te)
        
        window = np.zeros(self.ntimes, dtype=complex)
        w = np.where((self.times >= ts) & (self.times < te))
        if callable(f):
            window[w] = f(times=self.times[w])
        else:
            window[w] = f
            
        return window
       
    def get_amplitudes(self, ef=None, er=None):
        '''
        Converts a filter-drive-strengh to a resonator-drive-strenght and 
        vice-versa.
        '''
        if self.filterDrive:
            t = (-1j*self.params.t_factor*self.params.G)
        else:
            t = 1/(-1j*self.params.t_factor*self.params.G)
        
        if er is None:
            er = ef*t
        elif ef is None:
            ef = er/t
        
        return er, ef
    
    
    def add_filter_depletion_pulses(self, ts, tm, te, f1=1, f2=1):
        '''
        Calculates a two-stage drive pulse to deplete the filter
        (using the Green's fct).
        
        ts :    float; start time of the pulse in secs
        tm :    float; end of the first pulse and start time of the second pulse in secs
        te :    float; end time of the second pulses in secs
        f1 :    Amplitude specification for the first window:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        f2 :    See f1, just for the second window.
        '''
        self.add_filter_match_pulses(ts=ts, tm=tm, te=te, f1=f1, f2=f2, d=-1)
        
    def add_filter_match_pulses(self, ts, tm, te, d=-1, f1=1, f2=1):
        '''
        Calculates a two-stage drive pulse to achieve d times the field at
        ts in the filter.
        
        ts :    float; start time of the pulse in secs
        tm :    float; end of the first pulse and start time of the second pulse in secs
        te :    float; end time of the second pulses in secs
        d  :    float; factor of photon field at ts and te
        f1 :    Amplitude specification for the first window:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        f2 :    See f1, just for the second window.
        '''
        self.significantTimes.append(ts)
        self.significantTimes.append(tm)
        self.significantTimes.append(te)   
        
        window0 = self.get_drive_amplitude()
        window0[self.times >= ts] = 0
        
        window1 = self._new_window(ts=ts, te=tm, f=f1)
        window2 = self._new_window(ts=tm, te=te, f=f2)
        
        #solve linear equation to find depletion parameters
        A = np.array([
            [np.dot(window1, self.greens_function_alpha0[::-1]),
             np.dot(window2, self.greens_function_alpha0[::-1])], 
            [np.dot(window1, self.greens_function_alpha1[::-1]), 
             np.dot(window2, self.greens_function_alpha1[::-1])],
        ])

        b = np.array([
            np.dot(window0, self.greens_function_alpha0[::-1]),
            np.dot(window0, self.greens_function_alpha1[::-1]), 
        ])
	
        "depletion parameters"
        e1, e2 = np.linalg.solve(A, d*b)
        newWindow = e1*window1 + e2*window2
        
        self.windows.append(newWindow)
        
    def add_intercavity_depletion_pulses(self, ts, tm, te, f1=1, f2=1):
        '''
        Calculates a two-stage drive pulse to deplete the readout-resonator
        (using the Green's fct).
        
        ts :    float; start time of the pulse in secs
        tm :    float; end of the first pulse and start time of the second pulse in secs
        te :    float; end time of the second pulses in secs
        f1 :    Amplitude specification for the first window:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        f2 :    See f1, just for the second window.
        '''
        self.add_intercavity_match_pulses(ts=ts, tm=tm, te=te, f1=f1, f2=f2, d=-1)
        
    def add_intercavity_match_pulses(self, ts, tm, te, d=-1, f1=1, f2=1):
        '''
        Calculates a two-stage drive pulse to achieve d times the field at
        ts in the readout-resonator.
        
        ts :    float; start time of the pulse in secs
        tm :    float; end of the first pulse and start time of the second pulse in secs
        te :    float; end time of the second pulses in secs
        d  :    float; factor of photon field at ts and te
        f1 :    Amplitude specification for the first window:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        f2 :    See f1, just for the second window.
        '''
        
        window0 = self.get_drive_amplitude()
        window0[self.times >= ts] = 0
        
        window1 = self._new_window(ts=ts, te=tm, f=f1)
        window2 = self._new_window(ts=tm, te=te, f=f2)
        
        #solve linear equation to find depletion parameters
        A = np.array([
            [np.dot(window1, self.greens_function_beta0[::-1]),
             np.dot(window2, self.greens_function_beta0[::-1])], 
            [np.dot(window1, self.greens_function_beta1[::-1]),
             np.dot(window2, self.greens_function_beta1[::-1])],
        ])

        b = np.array([
            np.dot(window0, self.greens_function_beta0[::-1]),
            np.dot(window0, self.greens_function_beta1[::-1]), 
        ])
	
        "window amplitudes"
        e1, e2 = np.linalg.solve(A, d*b)
        newWindow = e1*window1 + e2*window2
        
        self.windows.append(newWindow)
    
    def add_full_depletion_pulses(self, ts, t1, t2, t3, te, f1=1, f2=1, f3=1, f4=1):
        '''
        Calculates a four-stage drive pulse to deplete the filter
        and the readout-resonator
        
        ts :    float; start time of the pulse in secs
        t1-t3 : float; end of the previous and start of the next pulse window
        te :    float; end time of the pulses in secs
        f1-f1 : Amplitude specification for the according windows:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        '''
        self.add_full_match_pulses(ts=ts, t1=t1, t2=t2, t3=t3, d=-1, te=te, f1=f1, f2=f2, f3=f3, f4=f4)
    
    def add_full_match_pulses(self, ts, t1, t2, t3, te, d=-1, f1=1, f2=1, f3=1, f4=1):
        '''
        Calculates a four-stage drive pulse to achieve d times the field
        at time ts at time te in the filter AND the readout-resonator.
        
        ts :    float; start time of the pulse in secs
        t1-t3 : float; end of the previous and start of the next pulse window
        te :    float; end time of the pulses in secs
        d  :    float; factor of photon field at ts and te
        f1-f1 : Amplitude specification for the according windows:
                float; uniform amplitude OR
                callable; accepting an array of times and returning the according
                    amplitudes (=pulse shape)
        '''
        self.significantTimes.append(ts)
        self.significantTimes.append(t1)
        self.significantTimes.append(t2)
        self.significantTimes.append(te)
        
        window0 = self.get_drive_amplitude()
        window0[self.times >= ts] = 0
        
        window1 = self._new_window(ts=ts, te=t1, f=f1)
        window2 = self._new_window(ts=t1, te=t2, f=f2)
        window3 = self._new_window(ts=t2, te=t3, f=f3)
        window4 = self._new_window(ts=t3, te=te, f=f4)
        
        #solve linear equation to find depletion parameters
        A = np.array([
            [np.dot(window1, self.greens_function_beta0[::-1]),
             np.dot(window2, self.greens_function_beta0[::-1]),
             np.dot(window3, self.greens_function_beta0[::-1]),
             np.dot(window4, self.greens_function_beta0[::-1])], 
            [np.dot(window1, self.greens_function_beta1[::-1]), 
             np.dot(window2, self.greens_function_beta1[::-1]), 
             np.dot(window3, self.greens_function_beta1[::-1]), 
             np.dot(window4, self.greens_function_beta1[::-1])],
            [np.dot(window1, self.greens_function_alpha0[::-1]),
             np.dot(window2, self.greens_function_alpha0[::-1]),
             np.dot(window3, self.greens_function_alpha0[::-1]),
             np.dot(window4, self.greens_function_alpha0[::-1])], 
            [np.dot(window1, self.greens_function_alpha1[::-1]), 
             np.dot(window2, self.greens_function_alpha1[::-1]), 
             np.dot(window3, self.greens_function_alpha1[::-1]), 
             np.dot(window4, self.greens_function_alpha1[::-1])],
        ])

        b = np.array([
            np.dot(window0, self.greens_function_beta0[::-1]), 
            np.dot(window0, self.greens_function_beta1[::-1]), 
            np.dot(window0, self.greens_function_alpha0[::-1]),
            np.dot(window0, self.greens_function_alpha1[::-1]), 
        ])
	
        "pulse amplitudes"
        e1, e2, e3, e4 = np.linalg.solve(A, d*b)
        newWindow = e1*window1 + e2*window2 + e3*window3 + e4*window4
        
        self.windows.append(newWindow)
        
    
    def get_drive_amplitude(self, er=1):
        '''
        Gives the amplitudes of the drive signal (envelope) an an array of length len(times).
        
        er : float; Drive amplitude factor
        '''
        return np.sum(self.windows, axis=0)*er
    
    def get_actual_drive(self, er=1):
        '''
        Gives the actual signal as an array of length len(times).
        
        er : float; Drive amplitude factor
        '''
        return self.get_drive_amplitude(er=er)*np.exp(1j*self.times*self.params.w_d)
    
    def get_drive_PSD(self, er=1, window=False, normalize=False):
        '''
        Gives the power spectral density of the drive signal.
        
        er : float; Drive amplitude factor
        window : boolean; shall we use a windowing fct?
        normalize: boolean; shall we normalize the output?
        '''
        dt = self.params.dt
        sig = self.get_actual_drive(er=er)
        
        return get_PSD(times=self.times, sig=sig, dt=dt, window=window, normalize=normalize)
    
    def plot_drive_amplitude(self, er=1, imag=False):
        p = self.get_drive_amplitude(er=er)
        
        for t in self.significantTimes:
            mi = min(min(p.real), min(p.imag))
            ma = max(max(p.real), max(p.imag))
            d = max(abs(mi), abs(ma))*0.1
            plt.plot([t*1e9, t*1e9], [ma+d, mi-d], color="lightgray")
        
        plt.plot(self.times*1e9, p.real)
        if imag:
            plt.plot(self.times*1e9, p.imag, '--')
        
        plt.xlabel(xlabel="Time [ns]")
        plt.ylabel(ylabel="Drive Pulse Amplitude [a.u.]")
        
    def plot_actual_drive(self, er=1, imag=False):
        p = self.get_actual_drive(er=er)
        
        for t in self.significantTimes:
            mi = min(min(p.real), min(p.imag))
            ma = max(max(p.real), max(p.imag))
            d = max(abs(mi), abs(ma))*0.1
            plt.plot([t*1e9, t*1e9], [ma+d, mi-d], color="lightgray")
        
        plt.plot(self.times*1e9, p.real)
        if imag:
            plt.plot(self.times*1e9, p.imag, '--')
        
        plt.xlabel(xlabel="Time [ns]")
        plt.ylabel(ylabel="Drive Amplitude [a.u.]")
        
    
    def get_dephasing_shared(self, filterField, filterTimes):
        dt = filterTimes[1]-filterTimes[0]
        greens_function_alpha0, greens_function_alpha1, _, _ = self._simple_greensFunctions(times=filterTimes)
        
        resp0 = dt*np.convolve(greens_function_alpha0, filterField)[:len(filterTimes)]
        resp1 = dt*np.convolve(greens_function_alpha1, filterField)[:len(filterTimes)]
        
        return self._field2deph(resp0=resp0, resp1=resp1, dt=dt)
    
    def get_dephasing_dedicated(self, er=1):
        greens_function_alpha0, greens_function_alpha1, _, _ = self._full_greensFunctions(times=self.times)
        
        filterField = self.get_drive_amplitude(er=er)
        resp0 = self.params.dt*np.convolve(greens_function_alpha0, filterField)[:self.ntimes]
        resp1 = self.params.dt*np.convolve(greens_function_alpha1, filterField)[:self.ntimes]
        
        return self._field2deph(resp0=resp0, resp1=resp1, dt=self.params.dt)
        
    def _field2deph(self, resp0, resp1, dt):
        d_gm = 2*abs(self.params.chi)*(resp0*resp1.conjugate()).imag
        gm = dt*np.cumsum((d_gm))
        
        return d_gm, gm
    
#####################################################################################
class PhotonFieldSolver:
    
    def __init__(self, qubit, numericSim=False):
        '''
        This calculates:
        + inter-readout-cavity field for 0 and 1 state:
            self.resp_alpha0, self.resp_alpha1
        + inter-filter-cavity field for 0 and 1 state:
            self.resp_beta0, self.resp_beta1
        
        
        Arguments:
            qubit :     A PurcellFilterQubit instance
            numericSim: Should we do a full numeric simulation or use the Green's
                        function?
        '''
        filterDrive = qubit.filterDrive
        fullSim = qubit.fullSim
        self.qubit = qubit
        self.params = qubit.params
        self.edrive = qubit.get_drive_amplitude()
        self.times = qubit.times
        self.ntimes = qubit.ntimes
        
        #Drive amplitudes
        if filterDrive:
            ers, efs = self.qubit.get_amplitudes(ef=self.edrive)
        else:
            ers, efs = self.qubit.get_amplitudes(er=self.edrive)

        
        
        ####################################################################################
        if numericSim:        
            #Do a full numeric coupled systems simulation
            
            #Constant parameters of the system
            params0 = [self.qubit.params.w_d, self.qubit.params.w_r,
                       self.qubit.params.w_f, self.qubit.params.chi,
                       self.qubit.params.G, self.qubit.params.kappa_f,
                       self.qubit.params.kappa_eff, self.qubit.params.dt]
            
            #Change parameters for directly or indirectly driven readout
            if self.qubit.filterDrive:
                params0.append(efs)
                vf = vectorfields_filterDrive
            else:
                params0.append(ers)
                vf = vectorfields_resonatorDrive
            
            
            params1 = params0[:] #Slice to copy!
            params1[3] *= -1 #Invert chi
            
            #self._numericSolverScipy(params0=params0, params1=params1, vf=vf) #This seems still to struggle with abritrary wave-forms!
            self._numericSolverSelf(params0=params0, params1=params1, vf=vf) #Best, but slow solution for now.
            #self._numericSolverSelfOpti(params0=params0, params1=params1, vf=vf) #WARNING: This is dirty & deprecated!
            
        ####################################################################################
        else:
            if filterDrive: #Are we driving the readout- or filter-cavity?
                #Use the greens function to find the filter-resonator-field
                if fullSim:
                    resDrive = 1j*efs
                else:
                    resDrive = ers
            else: #resonator drive
                if fullSim:
                    resDrive = 1j*ers
                else:
                    resDrive = ers
            
            #Do a simpler simulation, where we use the Greens functions to find a solution
            self.resp_alpha0 = self.params.dt*np.convolve(self.qubit.greens_function_alpha0, resDrive)[:self.qubit.ntimes]
            self.resp_alpha1 = self.params.dt*np.convolve(self.qubit.greens_function_alpha1, resDrive)[:self.qubit.ntimes]
            
            #Drive from the readout-resonator, as seen by the filter
            p0 = np.array(self.params.Gstar*self.resp_alpha0)
            p1 = np.array(self.params.Gstar*self.resp_alpha1)
            
            if filterDrive: #Are we driving the readout- or filter-cavity?
                #Use the greens function to find the filter-resonator-field
                if fullSim:
                    #If we do the full sim, we drive the filter only with
                    #the direct drive, the Greens function already consideres
                    #the indirect drive throught the readout-resonator
                    filDrive0 = 1j*efs
                    filDrive1 = 1j*efs
                else:
                    #In the simplified simulation, we have to emulate the drive
                    #by the filter
                    filDrive0 = p0 + efs
                    filDrive1 = p1 + efs
            else: #Resonator drive
                if fullSim:
                    filDrive0 = 1j*ers
                    filDrive1 = 1j*ers
                else:
                    #We drive the readout resonator, so the Greens function for
                    #the filter only 'sees' a drive from coupling with readout-resonator.
                    filDrive0 = p0
                    filDrive1 = p1
                    
            self.resp_beta0 = self.params.dt*np.convolve(self.qubit.greens_function_beta0, filDrive0)[:self.qubit.ntimes]
            self.resp_beta1 = self.params.dt*np.convolve(self.qubit.greens_function_beta1, filDrive1)[:self.qubit.ntimes]
        
    

    def get_intercavity_fields(self, er=1):
        return er*self.resp_alpha0, er*self.resp_alpha1
    
    
    def get_filter_fields(self,  er=1):
        if er is None:
            er = 1
        return np.array(er*self.resp_beta0), np.array(er*self.resp_beta1)
    
    def plot_alphas_time(self, er=1, plot_imag=True):
        r0, r1 = self.get_intercavity_fields(er=er)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\alpha_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\alpha_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\alpha_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\alpha_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()
    
    def plot_betas_time(self, er=None, plot_imag=True):
        r0, r1 = self.get_filter_fields(er=er)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\beta_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\beta_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\beta_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\beta_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()

    def get_filter_PSD(self, er=1, window=False, normalize=False):
        dt = self.params.dt
        sig = self.get_filter_fields(er=er)
        
        return get_PSD(times=self.times, sig=sig, dt=dt, window=window, normalize=normalize)
    
    def get_intercavity_PSD(self, er=1, window=False, normalize=False):
        dt = self.params.dt
        sig = self.get_intercavity_fields(er=er)
        
        return get_PSD(times=self.times, sig=sig, dt=dt, window=window, normalize=normalize)

    def plot_alphas_complex(self, er=None):
        a0, a1 = self.get_intercavity_fields(er=er)

        x0 = np.real(a0)
        y0 = np.imag(a0)
        x1 = np.real(a1)
        y1 = np.imag(a1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.1
        xd = max(abs(xmin), abs(xmax))*0.1
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\alpha_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\alpha_1$")
        
        plt.xlabel(xlabel="Real")
        plt.ylabel(ylabel="Imaginary")
        plt.legend()


    def plot_betas_complex(self, er=None):
        b0, b1 = self.get_filter_fields(er=er)
        
        x0 = np.real(b0)
        y0 = np.imag(b0)
        x1 = np.real(b1)
        y1 = np.imag(b1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.4
        xd = max(abs(xmin), abs(xmax))*0.4
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\beta_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\beta_1$")
        
        plt.xlabel(xlabel="Real")
        plt.ylabel(ylabel="Imaginary")
        plt.legend()
    

    def plot_optimal_weight_function(self, e0=1, iq=False):
        r0, r1 = self.get_intercavity_fields(er=e0)

        weight_function = r1 - r0
        self.weight_function_I = weight_function.real
        self.weight_function_Q = weight_function.imag
            
        if iq:
            plt.plot(weight_function.real, weight_function.imag)
            plt.xlabel(r"Re($\alpha_e - \alpha_g$)")
            plt.ylabel(r"Im($\alpha_e - \alpha_g$)")
        else:
            plt.plot(self.times, weight_function.real, label=r"Re($\alpha_e - \alpha_g$)")
            plt.plot(self.times, weight_function.imag, label=r"Im($\alpha_e - \alpha_g$)")
            plt.xlabel("Time [ns]")
            plt.legend()
    
    def get_SNR_time(self, er=1, eta=1, weightFct='optimal'):
        '''
        Set er OR ef for resonator drive or filter drive amplitude
        eta = Quantum efficiency (<=1)
        weightFct =     One of the following:
                        -'optimal' for the optimal weight function,
                        -a function working on the time array, returning
                        two arrarys describing the weight functions for I and Q,
                        -a Number being an even weight for I, Q and all times
                        -a two-touple being a weight for I and Q repectively.
        '''
        f0, f1 = self.get_filter_fields(er=er)
        
        kappa = self.params.kappa_f
        a = np.sqrt(2 * kappa * eta)
        diff = f0-f1
        v_I = (np.real(diff))
        v_Q = (np.imag(diff))
        
        if weightFct == 'optimal':
            weight_I = v_I#/np.sum(v_I**2*self.params.dt)
            weight_Q = v_Q#/np.sum(v_Q**2*self.params.dt)
            #print('Optimal')
        elif callable(weightFct):
            weight_I, weight_Q = weightFct(self.times)
            #print('User-Fct')
        elif isArray(weightFct) and len(weightFct) == 2:
            weight_I = np.ones_like(diff, dtype=float) * weightFct[0]
            weight_Q = np.ones_like(diff, dtype=float) * weightFct[1]
            #print('Two factors')
        else:
            weight_I = np.ones_like(diff, dtype=float) * weightFct
            weight_Q = np.ones_like(diff, dtype=float) * weightFct
            #print('Single factor')
            
        d_sig_t = a*((v_I*weight_I) + (v_Q*weight_Q)) * self.params.dt
        d_noise_t = ((weight_I**2) + (weight_Q**2)) * self.params.dt
        
        sig_t = np.cumsum(d_sig_t)
        noise_t_squared = np.cumsum(d_noise_t)
        noise_t = np.sqrt(noise_t_squared)
        
        SNR = np.zeros_like(self.times, dtype=float)
        d = np.where(noise_t != 0)
        SNR[d] = (sig_t[d] / noise_t[d])
        
        dSNR = np.zeros_like(self.times, dtype=float)
        dSNR[1:] = (SNR[1:] - SNR[:-1])/self.params.dt
        
        #SNR = np.cumsum(dSNR*self.params.dt) #Just for checkin'
        return (dSNR, SNR)
        
    
    def plot_SNR_time(self, er=1, eta=1, naiive=False):
        dSNR, SNR = self.get_SNR_time(er=er, eta=eta)
        
        #fig=plt.figure(figsize=(9,4))
        fig, ax1 = plt.subplots(figsize=(9,4))
        ax1.plot(self.times*1e9, SNR, "-", color='red', label="Optimal SNR")
        ax1.set_xlabel('Time [ns]')
        ax1.set_ylabel('SNR')
        
        ax2 = plt.twinx()
        ax2.plot(self.times*1e9, dSNR, "--", color='red', label=r"Optimal $\frac{d SNR}{dt}$")
        ax2.set_ylabel(ylabel=r"$\frac{d SNR}{dt}$")
        
        if naiive:    
            dSNRn, SNRn = self.get_SNR_time(er=er, eta=eta, weightFct=(1,1))
            ax1.plot(self.times*1e9, SNRn, "-", color='blue', label="Naïve SNR")
            ax2.plot(self.times*1e9, dSNRn, "--", color='blue', label=r"Naïve  $\frac{d SNR}{dt}$")
        
        ax1.legend()
        ax2.legend()
        #plt.legend()
        
    def get_fidelity_time(self, er=1, eta=1, weightFct='optimal'):
        dSNR, SNR = self.get_SNR_time(er=er, eta=eta, weightFct=weightFct)
        return (1-(erfc(SNR/(2*np.sqrt(2)))/2))
        
    def plot_fidelity_time(self, er=1, eta=1, naiive=False):
        fid = self.get_fidelity_time(er=er, eta=eta)
        
        if naiive:
            fidn = self.get_fidelity_time(er=er, eta=eta, weightFct=(1,1))
            plt.plot(self.times*1e9, fidn, "-", color='blue', label="Naïve SNR")
            plt.plot(self.times*1e9, fid, "-", color='red', label="Optimal SNR")
            plt.legend()
        else:
            plt.plot(self.times*1e9, fid, "-")
        
        #plt.yscale('log')
        plt.xlabel('Time [ns]')
        plt.ylabel('F')
        
    def plot_infidelity_time(self, er=1, eta=1, naiive=False):
        fid = self.get_fidelity_time(er=er, eta=eta)
        
        if naiive:
            fidn = self.get_fidelity_time(er=er, eta=eta, weightFct=(1,1))
            plt.plot(self.times*1e9, fidn, "-", color='blue', label="Naïve SNR")
            plt.plot(self.times*1e9, fid, "-", color='red', label="Optimal SNR")
            plt.legend()
        else:
            plt.plot(self.times*1e9, 1-fid, "-")
        
        plt.yscale('log')
        plt.xlabel('Time [ns]')
        plt.ylabel('1-F')
    
    def getOptimizePhotons(self, n=1, method='max-either', filt=False):
        if filt:
            f0, f1 = self.get_filter_fields(er=1)
        else:
            f0, f1 = self.get_intercavity_fields(er=1)
        n0 = np.abs(f0)**2
        n1 = np.abs(f1)**2

        method = method.lower()
        if method == 'max-either':
            nc = max(np.max(n0), np.max(n1))
        elif method == 'max-0':
            nc = max(n0)
        elif method == 'max-1':
            nc = max(n1)
        elif method == 'average-either':
            nc = (np.average(n0) + np.average(n1))/2
        elif method == 'average-1':
            nc = np.average(n1)
        elif method == 'average-0':
            nc = np.average(n0)
        else:
            raise NotImplementedError("Unknown Optimization Method: '" + method + "'. "+
                                      "Available are average-x and max-x, where x can be 0, 1 or 'either'.")
        
        er = np.sqrt(n/nc)
        return er
    
    def _numericSolverScipy(self, params0, params1, vf):
        w0 = [0, 0] #Both resonators start empty
        
        from scipy.integrate import ode
        #Create the ODE and call the ODE solver for the two possible canonic Qubit states
        diffeq0 = ode(vf).set_integrator(name='zvode', method='bdf', with_jacobian=False)
        diffeq0.set_initial_value(w0)
        diffeq0.set_f_params(params0)
        
        self.resp_alpha0, self.resp_beta0 = solve_ODE_scipy(qubit=self.qubit, diffeq=diffeq0)
        del(diffeq0)
        
        diffeq1 = ode(vf).set_integrator(name='zvode', method='bdf', with_jacobian=False)
        diffeq1.set_initial_value(w0)
        diffeq1.set_f_params(params1)
        
        self.resp_alpha1, self.resp_beta1 = solve_ODE_scipy(qubit=self.qubit, diffeq=diffeq1)
        del(diffeq1)
        
        
    def _numericSolverSelf(self, params0, params1, vf):
        self.resp_alpha0, self.resp_beta0 = solve_ODE_self(qubit=self.qubit, func=vf, par=params0)
        self.resp_alpha1, self.resp_beta1 = solve_ODE_self(qubit=self.qubit, func=vf, par=params1)
        
    def _numericSolverSelfOpti(self, params0, params1, vf):  #WARNING: This is dirty & deprecated!
        #TODO: This is still dirty! edrive should not be ovwritten (what about qubit.edrive?)
        #and we're solving this for the two states seperately, truncating one of the solutions...
        drive, self.resp_alpha0, self.resp_beta0 = solve_ODE_peak_and_hold(qubit=self.qubit, func=vf, par=params0)
        drive, self.resp_alpha1, self.resp_beta1 = solve_ODE_peak_and_hold(qubit=self.qubit, func=vf, par=params1)
        #or
        #drive, self.resp_alpha0, self.resp_beta0 = solve_ODE_beat_up_and_down(qubit=self.qubit, func=vf, par=params0)
        #drive, self.resp_alpha1, self.resp_beta1 = solve_ODE_beat_up_and_down(qubit=self.qubit, func=vf, par=params1)
        self.edrive = drive


#####################################################################################
def avPhNum(f):
    n = np.average(np.abs(f)**2)
    return n

def isArray(a):
    return hasattr(a, '__len__')

def get_PSD(times, sig, dt, window=False, normalize=False):
    N = times.size
    
    if window:
        from scipy.signal import blackman
        win = blackman(N)
    else:
        win = 1
    
    
    fr = np.fft.fftfreq(N, d=dt)
    w = np.where(fr >= 0) #Only use positive frequency plane
    freq = fr[w]
    
    ft = np.fft.fft(sig*win)[w]
    
    if normalize:
        ft = ft/max(abs(ft))
    
    return freq, ft
    
#####################################################################################


def vectorfields_filterDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func]
    """
    alpha, beta = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(ef_func):
        ef = ef_func(t)
    elif isArray(ef_func):    
        l=len(ef_func)
        if  (l*1.01) > i:
            ef = ef_func[min(i,l-1)]
        else:
            ef = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        ef = ef_func
        
    #print(i, ef)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta
    db = - 1j*(w_f-w_d)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta - 1j*ef
    
    return [da, db]


def vectorfields_resonatorDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta, i]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func]
    """
    alpha, beta  = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(er_func):
        er = er_func(t)
    elif isArray(er_func):
        l=len(er_func)
        if  (l*1.01) > i:
            er = er_func[min(i,l-1)]
        else:
            er = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        er = er_func
    
    #print(i, er)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta - 1j*er
    db = - 1j*(w_f-w_d)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta
    
    return [da, db]

def solve_ODE_scipy(qubit, diffeq):
    
    resp_alpha = np.zeros_like(qubit.times, dtype=complex)
    resp_beta = np.zeros_like(qubit.times, dtype=complex)
    t=0
    i=0
    while (diffeq.successful() and ((diffeq.t+qubit.params.dt) < qubit.params.tmax)):
        i += 1
        t = qubit.times[i]
        diffeq.integrate(t)
        a = complex(diffeq.y[0])
        b = complex(diffeq.y[1])
        #print(str(i)+"/"+str(len(qubit.times)), t, a, b)
        resp_alpha[i] = a
        resp_beta[i] = b
    
    return (resp_alpha, resp_beta)

def solve_ODE_self(qubit, func, par):
    resp_alpha = np.zeros_like(qubit.times, dtype=complex)
    resp_beta = np.zeros_like(qubit.times, dtype=complex)
    drive = par[-1] #Drive data array
    par[-1] = 0 #Will now be set one by one.
    
    t=0
    a=0
    b=0
    dt = qubit.params.dt
    for i, t in enumerate(qubit.times):
        par[-1] = drive[i]
        
        da, db = func(t=t, y=(a,b), par=par)
        
        resp_alpha[i] = a
        resp_beta[i] = b
        a += complex(da)*dt
        b += complex(db)*dt
        #dal, dbl = (da, db)
    
    return (resp_alpha, resp_beta)



def solve_ODE_beat_up_and_down(qubit, func, par): #WARNING: This is dirty & deprecated!
    drive = np.zeros_like(qubit.times, dtype=complex)
    resp = np.zeros_like([qubit.times]*2, dtype=complex)
    change = np.zeros_like(drive, dtype=complex)
    #drive = par[-1] #Drive data array
    dri=0
    par[-1] = dri #Will now be determined!
    
    t=0
    dt = qubit.params.dt
    
    rampTime = 2.48e-9
    
    kr = 1/qubit.params.kappa_r
    #kf = 1/qubit.params.kappa_f
    #g = 1/qubit.params.g
    #G = 1/qubit.params.G
    #kappa = max(kr, kf, g, G)
    kappa=(kr)
    
    f = tuple((0j,0j))
    d = tuple((0j,0j))
    maxPh = 0
    maxPhFilter = 0
    newP = 0
    newA = 0
    maxi = True
    w = 0j
    up = False
    
    
    for i, t in enumerate(qubit.times):
        photons = np.abs(f[0])**2
        photonsFilter = np.abs(f[1])**2
        maxPhFilter = max(photonsFilter, maxPhFilter)
        if maxi:
            maxPh = max(photons, maxPh)
        amp = 0.5*np.sqrt(photons)/kappa
        change[i] = amp
        
        #if kf > kr:
        if t < 2*rampTime:
            dri = np.exp(-1/2*((t-rampTime)/(rampTime/4))**2)
        else:
            dri = 0
            r=photons/maxPh
            if maxi:
                if r < 0.95:
                    maxi = False
            else:
                w = (f[0]/abs(f[0]))
                
                if t < newP+2*rampTime:
                    dri = newA*np.exp(-1/2*((t-newP)/(rampTime/4))**2)
                
                if (photonsFilter/maxPhFilter) < 0.001:
                    if t < newP+2*rampTime:
                        _=0 #Do nothing :o
                    elif (up and (r < 0.999)):
                        direction = +1
                    elif (not(up) and (r > 0.2)):
                        direction = -1
                    elif (not(up) and (r < 0.1)):
                        direction = +1
                    elif (up and (r > 0.9)):
                        direction = -1
                
                    if direction == -1:
                        direction = 0
                        up = True
                        print("depl %4.f ns, %.3f"%(t*1e9, (photons/maxPh)))
                        newP = t + rampTime
                        A = min(np.sqrt(photons/maxPh), 1)
                        newA = np.conj(w)*A
                    elif direction == +1:
                        direction = 0
                        print("accu %4.f ns, %3.f"%(t*1e9, (photons/maxPh)))
                        up = False
                        newP = t + rampTime
                        A = np.sqrt((maxPh-photons)/maxPh)
                        newA = w*A
        #else:
        #    if t < rampTime*1.1:
        #        dri = np.exp(-5*((t-rampTime)/rampTime)**2)
        #    else:
        #        amp = 0.55*np.sqrt(photons)/kappa
        #        dri -= (dri-amp)*(dt/rampTime)
        #        #dri = maxAmp*np.cos(2*pi*(t-rampTime*1.1)/(kr+kf))**2
        #        #dri = maxAmp
        
        par[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        drive[i] = dri
        
        d = func(t=t, y=f, par=par)
        
        resp[0,i] = f[0]
        resp[1,i] = f[1]
        f = (f[0] + d[0]*dt, f[1] + d[1]*dt)
    
    return drive, resp[0,:], resp[1,:]



def solve_ODE_peak_and_hold(qubit, func, par): #WARNING: This is dirty & deprecated!
    drive = np.zeros_like(qubit.times, dtype=complex)
    resp = np.zeros_like([qubit.times]*2, dtype=complex)
    change = np.zeros_like(drive, dtype=complex)
    #drive = par[-1] #Drive data array
    dri=0
    par[-1] = dri #Will now be determined!
    
    t=0
    dt = qubit.params.dt
    
    rampTime = 2.48e-9
    
    kr = 1/qubit.params.kappa_r
    #kf = 1/qubit.params.kappa_f
    #g = 1/qubit.params.g
    #G = 1/qubit.params.G
    #kappa = max(kr, kf, g, G)
    kappa=(kr)
    
    f = tuple((0j,0j))
    d = tuple((0j,0j))
    #maxPh = 0
    
    
    for i, t in enumerate(qubit.times):
        photons = np.abs(f[0])**2
        #photonsFilter = np.abs(f[1])**2
        amp = 0.9*np.sqrt(photons)/kappa
        change[i] = amp
        
        if t < 1.1*rampTime:
            dri = np.exp(-1/2*((t-rampTime)/(rampTime/4))**2)
        else:
            dri -= 2*(dri-amp)*(dt/rampTime)
        #else:
        #    if t < rampTime*1.1:
        #        dri = np.exp(-5*((t-rampTime)/rampTime)**2)
        #    else:
        #        amp = 0.55*np.sqrt(photons)/kappa
        #        dri -= (dri-amp)*(dt/rampTime)
        #        #dri = maxAmp*np.cos(2*pi*(t-rampTime*1.1)/(kr+kf))**2
        #        #dri = maxAmp
        
        par[-1] = dri/(-1j*qubit.params.t_factor*qubit.params.G)
        drive[i] = dri
        
        d = func(t=t, y=f, par=par)
        
        resp[0,i] = f[0]
        resp[1,i] = f[1]
        f = (f[0] + d[0]*dt, f[1] + d[1]*dt)
    
    return drive, resp[0,:], resp[1,:]

