#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 13 16:16:17 2017

@author: rene
"""

import numpy as np
from numpy import pi
from IPython.core.display import display, HTML
import warnings


class PurcellSetupParameterSet:
    w_rE = None
    w_rG = None
    chi = None
    w_r = None
    w_q = None
    w_f = None
    Qf = None
    kappa_r = None
    kappa_f = None
    G = None
    Gstar = None #Complex conjugate of G
    kappa_q = None
    w_d = None
    delta_rd = None
    delta_qr = None
    delta_fd = None
    delta_fq = None
    delta_wr = None
    delta = None
    t_factor = None
    n_crit = None
    anharm = None
    g = None
    t_purcell = None
    kappa_eff = None
    
    def _comp(self, a, b, level=0.01):
        return (abs(a-b) <= abs(a*level))
    
    def fully_determined(self, final=True):
        d = [
                self.delta_fd is None,
                self.kappa_f is None,
                self.w_rE is None,
                self.w_rG is None,
                self.chi is None,
                self.w_r is None,
                self.w_q is None,
                self.w_f is None,
                self.Qf is None,
                self.kappa_r is None,
                self.G is None,
                self.kappa_q is None,
                self.w_d is None,
                self.delta_rd is None,
                self.delta_wr is None,
                self.delta_qr is None,
                self.delta is None,
                self.g is None,
                self.anharm is None,
                self.dt is None,
                self.tmax is None,
                self.t_purcell is None,
                self.kappa_eff is None,
            ]
        if final:
            s = ["delta_fd", "kappa_f", "w_rE", "w_rG", "chi",
                 "w_r", "w_q", "w_f", "Qf", "kappa_r", "G",
                "kappa_q", "w_d", "delta_rd", "delta_wr",
                "delta_qr", "delta", "g", "anharm",
                "dt", "tmax", "t_purcell", "kappa_eff"]
            
            errStr = ""
            for i, j in enumerate(d):
                if j:
                    errStr += "\n\tVariable not set: '"+s[i]+"'"
            
            if not(errStr == ""):
                raise NameError("Trying to create parameter set, but: "+errStr)
            
        if sum(d) == 0:
            #Set some additional parameters
            self.t_factor = 1/(self.kappa_f/2+1j*self.delta_fd)
            self.Gstar = np.conjugate(self.G)
            
            #Near Eqs. 3 in Walter 2017
            self.n_crit = abs(self.delta_qr/(2*self.g))**2
            self.delta_fq = self.w_f - self.w_q
        
        return sum(d) == 0
    
    def is_well_defined(self, printWarnings=True):
        if self.fully_determined(final=printWarnings):
            
            sanity = []
            sanityMeaning = []
            
            #Chose something smaller than the shortest characteristic time.
            dt = min(2*pi/self.kappa_r, 2*pi/self.kappa_f, 2*pi/self.g, 2*pi/self.G, self.tmax/50)/30
            sanityMeaning.append("The chosen dt might be too large!")
            sanity.append(self.dt <= dt)
            
            #Eq. A13 from Sete/Martinis, 2015 ; Eqs. 3 in Walter 2017
            f = self.anharm/(self.anharm+self.delta_qr)
            chi = (self.g**2/self.delta_qr)*f
            sanityMeaning.append("%.3f MHz = [ g²/𝛥_qr * 𝛿_𝛼/(𝛿_𝛼+𝛥_qr)] != 𝜒 = %.3f MHz" %(chi/(2e6*pi), self.chi/(2e6*pi)))
            sanity.append(self._comp(chi, self.chi))
            
            #delta_fq = self.w_f - self.w_q
            
            
            #eqs 31 from Sete/Martinis, 2015
            gamma = self.g**2*np.abs(self.G)**2*self.kappa_f/(self.delta_qr**2 * ((self.delta_fq-np.abs(self.G)**2/self.delta_qr)**2 + (self.kappa_f/2)**2))
            #gamma = abs((self.g**2*self.kappa_q)/self.delta_qr**2)
            
            sanityMeaning.append("%.3f µs = T_Purcell != (eqs 31, SETE 2015) = %.3f µs" % (self.t_purcell*1e6, (1/gamma)*1e6))
            sanity.append(self._comp(1/gamma, self.t_purcell, level=0.1))
            
            
            sanityMeaning.append("w_r + chi != w_rG")
            sanity.append(self._comp((self.w_r + self.chi), self.w_rG))
            sanityMeaning.append("w_r - chi != w_rE")
            sanity.append(self._comp((self.w_r - self.chi), self.w_rE))
            
            sanityMeaning.append("w_f/𝜅_f != Q_f")
            sanity.append(self._comp((self.w_f/self.kappa_f), self.Qf))
            
            #sanity[7] = self._comp(self.G, np.sqrt((self.kappa_f*self.kappa_r)*(1+(2*(self.w_r-self.w_f)/self.kappa_f)**2)/4))

            # Eq. 10 from Sete/Martinis, 2015
            kr = self.calcKappar()
            sanityMeaning.append("%.3f ns = 1/𝜅_r != Eqs. 10 of [Sete, 2015] == %.3f ns" % (1/self.kappa_r*1e9, 1/kr*1e9))
            sanity.append(self._comp(self.kappa_r, kr))
            # Eq. 10 from Sete/Martinis, 2015

            keff = self.calcKappaeff()
            sanityMeaning.append("%.3f ns = 1/𝜅_eff != Eqs. 10 of [Sete, 2015] == %.3f ns" % (1/self.kappa_eff*1e9, 1/keff*1e9))
            sanity.append(self._comp(self.kappa_eff, keff))
            
            #Eq. 11 from Sete/Martinis, 2015
            kq = self.calcKappaq()
            sanityMeaning.append("%.4f µs = 1/𝜅_q != Eqs. 11 of [Sete, 2015] == %.4f µs" % (1/self.kappa_q*1e6, 1/kq*1e6))
            sanity.append(self._comp(self.kappa_q, kq))
            
            #Definitions
            sanityMeaning.append("w_q-w_r != 𝛥_qr")
            sanity.append(self._comp(self.delta_qr, (self.w_q - self.w_r)))
            
            sanityMeaning.append("w_r-w_d != 𝛥_rd")
            sanity.append(self._comp(self.delta_rd, (self.w_r - self.w_d)))
            
            sanityMeaning.append("w_f-w_d != 𝛥_fd")
            sanity.append(self._comp(self.delta_fd, (self.w_f - self.w_d)))
            
            sanityMeaning.append("w_r != -𝛥_rd 𝜅_r / 𝜅_f")
            sanity.append(self._comp(self.delta_wr, (-(self.delta_fd/self.kappa_f) * self.kappa_r)))
            
            sanityMeaning.append("𝛥 != 𝛥_rd + w_r")
            sanity.append(self._comp(self.delta, (self.delta_rd + self.delta_wr)))
            
            sanityMeaning.append("𝜒 < 𝛿_𝛼  is not fulfilled!")
            sanity.append(abs(self.chi) < abs(self.anharm))
            
            sanityMeaning.append("n_crit >> 1  is not fulfilled (dispersive approximation not valid)!")
            sanity.append(self.n_crit/5 > 1)
            
            keff = 4*(self.G**2)/(self.kappa_f*(1+(2*(self.w_d-self.w_f)/self.kappa_f)**2))
            sanityMeaning.append("%.4f µs = 1/𝜅_eff != Eq. 10 from Sete/Martinis, 2015 == %.4f µs" % (1/self.kappa_eff*1e6, 1/keff*1e6))
            sanity.append(self._comp(self.kappa_eff, keff))
            
            #Conditions for coupling g
            sanityMeaning.append("g < min(|𝛥_qr + 𝛿_𝛼|, | 𝛥_qr - 𝛿_𝛼|)  is not fulfilled!")
            sanity.append(self.g < min(abs(self.delta_qr+self.anharm), abs(self.delta_qr-self.anharm)))
            
            sanityMeaning.append("g << |𝛥_qr|  is not fulfilled!")
            sanity.append(self.g < abs(self.delta_qr)*5)
            
            if printWarnings:
                for i, t in enumerate(sanity):
                    if not(t):
                        s = "Check %d failed. That is: "%(i)
                        warnings.warn(s+sanityMeaning[i])
            
            return (sum(sanity) == len(sanity))
        else:
            return False
    
    def printParams(self):
        #Print all the parameters in a small table
        if self.fully_determined():
            beta_f = -1j * self.t_factor * self.Gstar
            ls = (self.g**2/self.delta_qr)
            
            display(HTML("<table>"
                +"<tr><th>Variable</th><th></th><th>Value</th><th>Unit</th></tr>"
                +"<tr><td>(𝜔<sub>r</sub> ± 𝜒)/2𝜋</td><td>=</td><td>(%.5f ± %.5f)</td><td>GHz</td></tr>" % (((self.w_r*1e-9)/(2*pi)), ((self.chi)/(2e9*pi)))
                +"<tr><td>lamb shift</td><td>=</td><td>(%.5f)</td><td>MHz</td></tr>" % ((ls/(2e6*pi)))
                +"<tr><td>𝜔<sub>f</sub>/2𝜋</td><td>=</td><td>%+.4f</td><td>GHz</td></tr>" % ((self.w_f)/(2e9*pi))
                +"<tr><td>𝜔<sub>q</sub>/2𝜋</td><td>=</td><td>%+.4f</td><td>GHz</td></tr>" % ((self.w_q)/(2e9*pi))
                +"<tr><td>𝜔<sub>d</sub>/2𝜋</td><td>=</td><td>%+.5f</td><td>GHz</td></tr>" % ((self.w_d)/(2e9*pi))
                +"<tr><td>𝛥<sub>rd</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_rd)/(2e6*pi))
                +"<tr><td>𝛥<sub>fd</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_fd)/(2e6*pi))
                +"<tr><td>𝛥<sub>qr</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_qr)/(2e6*pi))
                +"<tr><td>𝛥/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta)/(2e6*pi))
                +"<tr><td>𝛿𝜔<sub>r</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_wr)/(2e6*pi))
                +"<tr><td>𝜅<sub>f</sub></td><td>=</td><td>%+.4f</td><td>MHz</td></tr>" % (self.kappa_f/(pi*2e6))
                +"<tr><td>𝜅<sub>r</sub></td><td>=</td><td>%+.4f</td><td>MHz</td></tr>" % (self.kappa_r/(pi*2e6))
                +"<tr><td>𝜅<sub>eff</sub></td><td>=</td><td>%+.4f</td><td>MHz</td></tr>" % (self.kappa_eff/(pi*2e6))
                +"<tr><td>𝜅<sup>-1</sup><sub>q</sub></td><td>=</td><td>%+.4f</td><td>µs</td></tr>" % (1e6/self.kappa_q)
                +"<tr><td>T<sub>Purcell</sub> </td><td>=</td><td>%.4f</td><td>µs</td></tr>" % (self.t_purcell*1e6)
                +"<tr><td>𝜅/|𝜒|<sub>r</sub></td><td>=</td><td>%+.4f</td><td></td></tr>" % (self.kappa_r/abs(self.chi))
                +"<tr><td>G/2𝜋</td><td>=</td><td>%+.2f</td><td>MHz</td></tr>" % (self.G/(2e6*pi))
                +"<tr><td>g/2𝜋</td><td>=</td><td>%+.2f</td><td>MHz</td></tr>" % (self.g/(2e6*pi))
                +"<tr><td>Q<sub>f</sub></td><td>=</td><td>%+.2f</td><td></td></tr>" % self.Qf
                +"<tr><td>n<sub>crit</sub></td><td>=</td><td>%.3f</td><td></td></tr>" % (self.n_crit)
                #+"<tr style='color:black; height: 5px;'><td colspan='4'>&nbsp;</td></tr>"
                +"<tr><td>T</td><td>=</td><td>%d</td><td>ns</td></tr>" % (self.tmax*1e9)
                +"<tr><td>dT</td><td>=</td><td>%.4f</td><td>ns</td></tr>" % (self.dt*1e9)
                +"<tr><td>F<sup>-1</sup>=𝜅<sub>r</sub>/𝜅<sub>q</sub></td><td>=</td><td>%.4f</td><td></td></tr>" % (self.kappa_r/self.kappa_q)
                +"<tr><td>T<sub>Purcell</sub> ∙ 𝜅<sub>r</sub></td><td>=</td><td>%.4f</td><td></td></tr>" % (self.kappa_r*self.t_purcell)
                +"<tr><td>𝛽 = a ∙ 𝛼 ; a</td><td>=</td><td>(%+f, %+f j)</td><td></td></tr>" % (beta_f.real, beta_f.imag)
                +"<tr><td>𝛿<sub>α</sub></td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.anharm)/(2e6*pi))
                +"</table>"))

    def calcKappar(self):
        # Eq. 10 from Sete/Martinis, 2015
        return 4 * abs(self.G) ** 2 / self.kappa_f * (1 / (1 + (2 * (self.w_r - self.w_f) / self.kappa_f) ** 2))

    def calcKappaq(self):
        # Eq. 11 from Sete/Martinis, 2015
        return 4 * abs(self.G) ** 2 / self.kappa_f * (1 / (1 + (2 * (self.w_q - self.w_f) / self.kappa_f) ** 2))

    def calcKappaeff(self):
        # Eq. 10 from Sete/Martinis, 2015
        return 4 * abs(self.G) ** 2 / self.kappa_f * (1 / (1 + (2 * (self.w_d - self.w_f) / self.kappa_f) ** 2))

    def setFrequencies(self, w_r, w_f, w_q):
        self.w_r = w_r
        self.w_f = w_f
        self.w_q = w_q
        self.delta_fq = self.w_f - self.w_q
        self.delta_fr= self.w_f - self.w_r
        self.delta_qr = self.w_q - self.w_r
    
    def setCenterDrive(self):
        if self.w_r and self.kappa_f and self.kappa_r and self.w_f:
            w_d = (self.w_r*self.kappa_f/self.kappa_r-self.w_f)/(self.kappa_f/self.kappa_r - 1)
            self.setDriveFreq(w_d=w_d)
        else:
            warnings.warn("w_r, kappa_f, kappa_r and w_f must be set to calculate the center-drive frequency!")
    
    def setDriveFreq(self, w_d):
        self.w_d = w_d
        self.delta_rd = self.w_r - w_d
        self.delta_fd = self.w_f - w_d
    
    def setChi(self, chi):
        self.chi = chi
        self.w_rG = self.w_r + self.chi
        self.w_rE = self.w_r - self.chi

    def init1(self, w_r, w_f, w_q, g, anharm, kappa_f, kappa_chi):
        self.setFrequencies(w_r=w_r, w_f=w_f, w_q=w_q)
        self.g = g
        self.kappa_f = kappa_f
        self.Qf = self.w_f/self.kappa_f
        self.anharm = anharm

        f = self.anharm/(self.anharm+self.delta_qr)
        self.setChi(self.g**2/self.delta_qr*f)

        self.kappa_eff = kappa_chi * abs(self.chi)
        self.G = np.sqrt(self.kappa_eff*(self.kappa_f**2 + 4*self.delta_fr**2)/(4*self.kappa_f))


        self.kappa_r = (4*abs(self.G)**2/self.kappa_f)*(1/(1+(2*(self.w_r-self.w_f)/self.kappa_f)**2))
        self.kappa_q = (4*abs(self.G)**2/self.kappa_f)*(1/(1+(2*(self.w_q-self.w_f)/self.kappa_f)**2))

        self.setCenterDrive()

        self.delta_wr = -(self.delta_fd/self.kappa_f) * self.kappa_r
        self.delta = self.w_r - self.w_d + self.delta_wr

        a=((self.delta_fq - abs(self.G)**2/self.delta_qr)**2 + (self.kappa_f/2)**2)
        b=(self.g**2*abs(self.G)**2*self.kappa_f)
        self.t_purcell = self.delta_qr**2 * a/b

    def init_meas(self, w_r, w_f, w_d, w_q, g, anharm, kappa_f, G):
        self.setFrequencies(w_r=w_r, w_f=w_f, w_q=w_q)
        self.g = g
        self.kappa_f = kappa_f
        self.Qf = self.w_f/self.kappa_f
        self.anharm = anharm

        f = self.anharm/(self.anharm+self.delta_qr)
        self.setChi(self.g**2/self.delta_qr*f)

        self.setDriveFreq(w_d)
        self.G = G
        self.kappa_eff = self.calcKappaeff()
        self.kappa_r = self.calcKappar()
        self.kappa_q = self.calcKappaq()

        self.delta_wr = -(self.delta_fd/self.kappa_f) * self.kappa_r
        self.delta = self.w_r - self.w_d + self.delta_wr

        a=((self.delta_fq - abs(self.G)**2/self.delta_qr)**2 + (self.kappa_f/2)**2)
        b=(self.g**2*abs(self.G)**2*self.kappa_f)
        self.t_purcell = self.delta_qr**2 * a/b

    def setTimes(self, tmax, secFac=1):
        self.tmax = tmax
        self.dt = tmax/1000
        if self.fully_determined(final=False):
            #Calculate dt
            G = self.G
            wa = np.array([self.w_r - self.w_d + self.chi, self.w_r - self.w_d - self.chi])
            wb = self.w_f - self.w_d - 1j * self.kappa_f/2
        
            s = np.sqrt((wa-wb)**2 + 4*np.abs(G)**2)
            wp = (wa + wb + s)/2
            wm = (wa + wb - s)/2
            
            self.dt = min(min(np.abs(2*pi/wa)), np.abs(2*pi/wb), min(np.abs(2*pi/wp)), min(np.abs(2*pi/wm)), self.tmax/50)/(secFac*200)
        else:
            self.tmax = None
            self.dt = None