#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt

class SquareDepletionMeasurementPulse:
    def __init__(self, delta_r, chi, kappa, t0, t1, tmax, dt, amps=None, depletion=True):
        """
        Calculate and inspect a measurement pulse on a linear resonator.
        
        Parameters
        
        Resonator:
            chi: cavity pull, angular frequency [rad/ns] = 2*pi*chi*1e9
            kappa: resonator decay constant [1/ns] (i.e. linewidth is 2*pi*kappa!)
            
        Pulse:
            delta_r: detuning from bare resonator frequency, [rad/ns]  = 2*pi*delta_r*1e9
            t0, t1: begin of first/second depletion pulse
            tmax: end of measurement pulse
        
            dt: time-step used for calculation. choose much less than 1/kappa, 1/chi, any t.
        """
        self.delta_r = delta_r
        self.chi = chi
        self.kappa = kappa
                
        self.dt = dt
        self.tmax = tmax
        self.times = np.arange(0, tmax, dt)
        
        self.ntimes = len(self.times)
        if amps is not None:
            self.e0 = amps[0]
	else:
	     self.e0 = 1

        
        #Green's function of the resonator
        
        self.greens_function_zero = -1j*np.exp((-1j*(delta_r + chi) - kappa/2)*self.times)
        self.greens_function_one = -1j*np.exp((-1j*(delta_r - chi) - kappa/2)*self.times)


        #components of measurement pulse
        
        self.window0 = np.zeros(self.ntimes)
        self.window0[np.where(self.times < t0)] = 1
        
        #solve linear equation to find depletion parameters
        
        "depletion pulse"
        if depletion:
            if amps is not None and len(amps) == 3:
		self.window1 = np.zeros(self.ntimes)
		self.window1[np.where((self.times >= t0) & (self.times < t1))] = 1
		self.window2 = np.zeros(self.ntimes)
		self.window2[np.where(t1 <= self.times)] = 1
		A = np.array([
			[np.dot(self.window1, self.greens_function_zero[::-1]),
			 np.dot(self.window2, self.greens_function_zero[::-1])], 
			[np.dot(self.window1, self.greens_function_one[::-1]), 
			 np.dot(self.window2, self.greens_function_one[::-1])],     
		])
	    
		b = np.array([
			np.dot(self.window0, self.greens_function_zero[::-1]),
			np.dot(self.window0, self.greens_function_one[::-1]), 
		])
		    
		"depletion parameters"
		self.e1, self.e2 = np.linalg.solve(A, -b)
	    else:
	        self.e1 = amps[1]
	        self.e2 = amps[2]
            self.edrive = self.e0 * self.window0 + self.e1*self.window1 + self.e2*self.window2
        else:
            self.edrive = self.e0 * self.window0

        "inter-cavity field for 0 and 1 state"
        self.resp0 = self.dt*np.convolve(self.greens_function_zero, self.edrive)[:self.ntimes]
        self.resp1 = self.dt*np.convolve(self.greens_function_one, self.edrive)[:self.ntimes]

    def get_intercavity_fields(self, e0=1):
        return e0*self.resp0, e0*self.resp1
    
    
    def plot_alphas(self, e0=1, plot_imag=True):
        
        r0, r1 = self.get_intercavity_fields(e0)
        
        plt.plot(self.times, r0.real, "r-", label=r"Re($\alpha_0$)")
        plt.plot(self.times, r1.real, "g-", label=r"Re($\alpha_1$)")
        if plot_imag:
            plt.plot(self.times, r0.imag, "r--", label=r"Im($\alpha_0$)")
            plt.plot(self.times, r1.imag, "g--", label=r"Im($\alpha_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()

    def plot_optimal_weight_function(self, e0=1, iq=False):
        
        r0, r1 = self.get_intercavity_fields()

        weight_function = r1 - r0
        self.weight_function_I=weight_function.real
        self.weight_function_Q=weight_function.imag
        
        if iq:
            plt.plot(weight_function.real, weight_function.imag)
            plt.xlabel(r"Re($\alpha_e - \alpha_g$)")
            plt.ylabel(r"Im($\alpha_e - \alpha_g$)")
        else:
            plt.plot(self.times, weight_function.real, label=r"Re($\alpha_e - \alpha_g$)")
            plt.plot(self.times, weight_function.imag, label=r"Im($\alpha_e - \alpha_g$)")
            plt.xlabel("Time [ns]")
            plt.legend()
            
    def optimized_snr_slope(self, eta=1.0):
        "Obtain slope a of SNR using optimized weight functions"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(e0=1)
        return 4*self.kappa*eta*np.sum(np.absolute(r0 - r1)**2)*self.dt
    
    def naive_weights_snr(self, eta=1.0, integration_factor=1):
        "obtain slope a of SNR using constant weight function (with optimal phase):"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(e0=1)
        sig = 2*np.sqrt(eta*self.kappa)*np.absolute(np.sum((r0 - r1)))*self.dt
        noise = np.sqrt(self.tmax*integration_factor)
        return (sig/noise)**2
        
    def dephasing_strength(self):
        "Obtain dephasing rate b as"
        "contrast = exp(-b * e0**2)"        
        r0, r1 = self.get_intercavity_fields()
        return 2*self.chi*np.sum(np.imag(r1 * r0.conj())*self.dt)
    
    def estimate_eta(self, real_eta=1.0,  weight_method='optimal', integration_factor=1):
        if weight_method == 'optimal':
            snr_a = self.optimized_snr_slope(real_eta)
        else:
            snr_a = self.naive_weights_snr(real_eta, integration_factor)
        deph = self.dephasing_strength()
        return snr_a/(8*deph)
