#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
import matplotlib.pyplot as plt
import warnings
from scipy.special import erfc

class PurcellFilterQubit:
    def __init__(self, params):
        params.is_well_defined()
        
        self.params = params
        
        self.times = np.arange(0, self.params.tmax, self.params.dt)
        self.ntimes = len(self.times)
        
        #self.pulse = np.zeros(self.ntimes)
        self.windows = []
        self.windows.append(np.zeros(self.ntimes))
        self.significantTimes=[]
        
        #Green's function of the resonator
        self.greens_function_alpha0 = -1j*np.exp((-1j*(self.params.delta + self.params.chi) - self.params.kappa_r/2)*self.times)
        self.greens_function_alpha1 = -1j*np.exp((-1j*(self.params.delta - self.params.chi) - self.params.kappa_r/2)*self.times)
        #Green's function of the filter
        self.greens_function_beta0 = -1j*np.exp((-1j*(self.params.delta_fd + self.params.chi) - self.params.kappa_f/2)*self.times)
        self.greens_function_beta1 = -1j*np.exp((-1j*(self.params.delta_fd - self.params.chi) -self. params.kappa_f/2)*self.times)
        
    def add_pulse(self, ts, te, f):
        self.significantTimes.append(ts)
        self.significantTimes.append(te)
        
        w = (self.times>=ts) * (self.times<te)
        newWindow = np.zeros(self.ntimes)
        if callable(f):
            newWindow[w] = f(times=self.times[w])
        else:
            newWindow[w] = f
        self.windows.append(newWindow)
        
    def add_depletion_pulses(self, ts, tm, te, f1, f2):  
        self.significantTimes.append(ts)
        self.significantTimes.append(tm)
        self.significantTimes.append(te)   
        
        window0 = self.get_pulse()
        window0[self.times < ts] = 0
        
        w1 = np.where((self.times >= ts) & (self.times < tm))
        self.window1 = np.zeros(self.ntimes)
        if callable(f1):
            self.window1[w1] = f1(times=self.times[w1])
        else:
            self.window1[w1] = f1
        
        
        w2 = np.where((self.times >= tm) & (self.times < te))
        self.window2 = np.zeros(self.ntimes)
        if callable(f2):
            self.window2[w2] = f2(times=self.times[w2])
        else:
            self.window2[w2] = f2
        #solve linear equation to find depletion parameters

        A = np.array([
            [np.dot(self.window1, self.qubit.greens_function_alpha0[::-1]),
             np.dot(self.window2, self.qubit.greens_function_alpha0[::-1])], 
            [np.dot(self.window1, self.qubit.greens_function_alpha1[::-1]), 
             np.dot(self.window2, self.qubit.greens_function_alpha1[::-1])],     
        ])

        b = np.array([
            np.dot(window0, self.qubit.greens_function_alpha0[::-1]),
            np.dot(window0, self.qubit.greens_function_alpha1[::-1]), 
        ])
	
        "depletion parameters"
        self.e1, self.e2 = np.linalg.solve(A, -b)
        newWindow = self.e1*self.window1 + self.e2*self.window2
        
        self.windows.append(newWindow)
        
    def get_pulse(self):
        return np.sum(self.windows, axis=0)
    
    def plot_pulse(self, er=1):
        p = self.get_pulse()
        
        for t in self.significantTimes:
            plt.plot([t*1e9, t*1e9], [max(p), min(p)], color="lightgray")
        
        plt.plot(self.times*1e9, p)
        
        plt.xlabel(s="Time [ns]")
        plt.ylabel(s="Drive Pulse Amplitude [a.u.]")
    
    
#####################################################################################
class PhotonFieldSolver:
    
    def __init__(self, qubit, filterDrive=False, fullSim=False, lazy=False):
        '''
        This calculates:
        inter-readout-cavity field for 0 and 1 state:
        self.resp_alpha0, self.resp_alpha1
        inter-filter-cavity field for 0 and 1 state:
        self.resp_beta0, self.resp_beta1
        
        
        Arguments:
            qubit :  a PurcellFilterQubit instance
            filterDrive :  Should the readout be driven through the readout line? (boolean)
            fullSim : Should we do a full or a simpler (faste?) simulation,
                        where we assume the filter to be much faster than the
                        readout-resonator (boolean)
            lazy :  (Only when fullSim=False): Should we be super lazy and
                    calculate the photon number in the filter as proportional
                    to the readout-resonator and drive or rather simulate the
                    filter with two drives: direct and readout-resonator? (boolean)
        
        '''
        self.qubit = qubit
        self.params = qubit.params
        self.edrive = qubit.get_pulse()
        self.times = qubit.times
        self.ntimes = qubit.ntimes
        
        
        "filter field for 0 and 1 state"
        ers = self.edrive
        efs = ers/(-1j*self.params.t_factor*self.params.G)
        
        if fullSim:
            #Do a full numeric coupled systems simulation
            from scipy.integrate import ode
            
            #Constant parameters of the system
            params0 = [self.qubit.params.w_d, self.qubit.params.w_r,
                       self.qubit.params.w_f, self.qubit.params.chi,
                       self.qubit.params.G, self.qubit.params.kappa_f,
                       self.qubit.params.kappa_r, self.qubit.params.dt]
            params1 = params0[:] #Slice to copy!
            params1[3] *= -1 #Invert chi
            w0 = [0, 0] #Both resonators start empty
            
            '''
            #Change parameters for directly or indirectly driven readout
            if filterDrive:
                #TODO: Use real pulse efs
                params0.append(1)
                params1.append(1)
                vf = vectorfields_filterDrive
            else:
                #TODO: Use real pulse ers
                params0.append(1)
                params1.append(1)
                vf = vectorfields_resonatorDrive
            
            #Create the ODE and call the ODE solver for the two possible canonic Qubit states
            diffeq0 = ode(vf).set_integrator(name='zvode', method='bdf', with_jacobian=False)
            diffeq0.set_initial_value(w0)
            diffeq0.set_f_params(params0)
            
            self.resp_alpha0, self.resp_beta0 = solve_ODE(qubit=self.qubit, diffeq=diffeq0)
            del(diffeq0)
            
            diffeq1 = ode(vf).set_integrator(name='zvode', method='bdf', with_jacobian=False)
            diffeq1.set_initial_value(w0)
            diffeq1.set_f_params(params1)
            
            self.resp_alpha1, self.resp_beta1 = solve_ODE(qubit=self.qubit, diffeq=diffeq1)
            del(diffeq1)
            '''
            
            #Change parameters for directly or indirectly driven readout
            if filterDrive:
                drivODE = efs
                vfODE = vectorfields_filterDrive
            else:
                drivODE = ers
                vfODE = vectorfields_resonatorDrive
                
            self.resp_alpha0, self.resp_beta0 = solve_ODE_self(qubit=qubit, func=vfODE, par=params0, drive=drivODE)
            self.resp_alpha1, self.resp_beta1 = solve_ODE_self(qubit=qubit, func=vfODE, par=params1, drive=drivODE)
            
        else:
            #Do a simpler simulation, where we assume the filter to be much
            #faster than the readout-resonator
            
            #Use the greens function to find the readout-resonator-field
            self.resp_alpha0 = self.params.dt*np.convolve(self.qubit.greens_function_alpha0, self.edrive)[:self.qubit.ntimes]
            self.resp_alpha1 = self.params.dt*np.convolve(self.qubit.greens_function_alpha1, self.edrive)[:self.qubit.ntimes]
            
            #Drive through readout-resonator, as seen by the filter
            p0 = np.array(self.params.Gstar*self.resp_alpha0)
            p1 = np.array(self.params.Gstar*self.resp_alpha1)
            
            if filterDrive: #Are we driving the readout- or filter-cavity?
                #Use the greens function to find the filter-resonator-field
                drive0 = p0 + efs
                drive1 = p1 + efs
                self.resp_beta0 = self.params.dt*np.convolve(self.qubit.greens_function_beta0, drive0)[:self.qubit.ntimes]
                self.resp_beta1 = self.params.dt*np.convolve(self.qubit.greens_function_beta1, drive1)[:self.qubit.ntimes]
            else:
                if lazy:
                    #Be lazy and calculate the photon number in the filter as
                    #proportional to the readout-resonator and drive
                    self.resp_beta0 = np.array(-1j*self.qubit.params.t_factor * self.params.Gstar * self.resp_alpha0)
                    self.resp_beta1 = np.array(-1j*self.qubit.params.t_factor * self.params.Gstar * self.resp_alpha1)
                else:
                    #Be a little less lazy and use the Greens function with two
                    #drives (direct and from coupling with readout-resonator)
                    #for the filter-photon-field
                    drive0 = p0
                    drive1 = p1
                    self.resp_beta0 = self.params.dt*np.convolve(self.qubit.greens_function_beta0, drive0)[:self.qubit.ntimes]
                    self.resp_beta1 = self.params.dt*np.convolve(self.qubit.greens_function_beta1, drive1)[:self.qubit.ntimes]
        
    

    def get_intercavity_fields(self, er=None, ef=None):
        er, ef =  self.get_amplitudes(ef=ef, er=er)
        return er*self.resp_alpha0, er*self.resp_alpha1
    
    def get_amplitudes(self, ef=None, er=None):
        if er is None:
            er = ef*(-1j*self.params.t_factor*self.params.G)
        elif ef is None:
            ef = er/(-1j*self.params.t_factor*self.params.G)
        return er, ef
    
    def get_filter_fields(self,  er=None, ef=None):
        er, ef = self.get_amplitudes(er=er, ef=ef)
        return np.array(er*self.resp_beta0), np.array(er*self.resp_beta1)
    
    def plot_alphas_time(self, er=None, ef=None, plot_imag=True):
        r0, r1 = self.get_intercavity_fields(er=er, ef=ef)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\alpha_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\alpha_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\alpha_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\alpha_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()
    
    def plot_betas_time(self, er=None, ef=None, plot_imag=True):
        r0, r1 = self.get_filter_fields(er=er, ef=ef)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\beta_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\beta_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\beta_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\beta_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()

    def plot_alphas_complex(self, er=None, ef=None):
        a0, a1 = self.get_intercavity_fields(er=er, ef=ef)

        x0 = np.real(a0)
        y0 = np.imag(a0)
        x1 = np.real(a1)
        y1 = np.imag(a1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.1
        xd = max(abs(xmin), abs(xmax))*0.1
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\alpha_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\alpha_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()


    def plot_betas_complex(self, er=None, ef=None):
        b0, b1 = self.get_filter_fields(er, ef)
        
        x0 = np.real(b0)
        y0 = np.imag(b0)
        x1 = np.real(b1)
        y1 = np.imag(b1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.4
        xd = max(abs(xmin), abs(xmax))*0.4
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\beta_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\beta_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()
    

    def plot_optimal_weight_function(self, e0=1, iq=False):
        r0, r1 = self.get_intercavity_fields(er=e0)

        weight_function = r1 - r0
        self.weight_function_I = weight_function.real
        self.weight_function_Q = weight_function.imag
            
        if iq:
            plt.plot(weight_function.real, weight_function.imag)
            plt.xlabel(r"Re($\alpha_e - \alpha_g$)")
            plt.ylabel(r"Im($\alpha_e - \alpha_g$)")
        else:
            plt.plot(self.times, weight_function.real, label=r"Re($\alpha_e - \alpha_g$)")
            plt.plot(self.times, weight_function.imag, label=r"Im($\alpha_e - \alpha_g$)")
            plt.xlabel("Time [ns]")
            plt.legend()
    
    def get_SNR_time(self, er=None, ef=None, nu=1, weightFct='optimal'):
        '''
        Set er OR ef for resonator drive or filter drive amplitude
        nu = Quantum efficieny (<=1)
        weightFct =     One of the following:
                        -'optimal' for the optimal weight function,
                        -a function working on the time array, returning
                        two arrarys describing the weight functions for I and Q,
                        -a Number being an even weight for I, Q and all times
                        -a two-touple being a weight for I and Q repectively.
        '''
        f0, f1 = self.get_filter_fields(er=er, ef=ef)
        
        a = 2 * np.sqrt(self.params.kappa_r * nu)
        diff = (f0 - f1)
        v_I = np.abs(np.real(diff))
        v_Q = np.abs(np.imag(diff))
        
        if weightFct == 'optimal':
            weight_I = v_I
            weight_Q = v_Q
            #print('Optimal')
        elif callable(weightFct):
            weight_I, weight_Q = weightFct(self.times)
            #print('User-Fct')
        elif isArray(weightFct) and len(weightFct) == 2:
            weight_I = np.ones_like(diff, dtype=float) * weightFct[0]
            weight_Q = np.ones_like(diff, dtype=float) * weightFct[1]
            #print('Two factors')
        else:
            weight_I = np.ones_like(diff, dtype=float) * weightFct
            weight_Q = np.ones_like(diff, dtype=float) * weightFct
            #print('Single factor')
            
        d_sig_t = a*((v_I*weight_I) + (v_Q*weight_Q)) * self.params.dt
        d_noise_t = ((weight_I**2) + (weight_Q**2)) * self.params.dt
        
        sig_t = np.cumsum(d_sig_t)
        noise_t_squared = np.cumsum(d_noise_t)
        noise_t = np.sqrt(noise_t_squared)
        
        SNR = np.zeros_like(self.times, dtype=float)
        d = np.where(noise_t != 0)
        SNR[d] = (sig_t[d] / noise_t[d])
        
        dSNR = np.zeros_like(self.times, dtype=float)
        dSNR[1:] = (SNR[1:] - SNR[:-1])/self.params.dt
        
        #SNR = np.cumsum(dSNR*self.params.dt) #Just for checkin'
        return (dSNR, SNR)
    
    def plot_SNR_time(self,er=None, ef=None, nu=1, naiive=False):
        dSNR, SNR = self.get_SNR_time(er=er, ef=ef, nu=nu)
        
        #fig=plt.figure(figsize=(9,4))
        fig, ax1 = plt.subplots(figsize=(9,4))
        ax1.plot(self.times*1e9, SNR, "-", color='red', label="Optimal SNR")
        ax1.set_xlabel('Time [ns]')
        ax1.set_ylabel('SNR')
        
        ax2 = plt.twinx()
        ax2.plot(self.times*1e9, dSNR, "--", color='red', label=r"Optimal $\frac{d SNR}{dt}$")
        ax2.set_ylabel(ylabel=r"$\frac{d SNR}{dt}$")
        
        if naiive:    
            dSNRn, SNRn = self.get_SNR_time(er=er, ef=ef, nu=nu, weightFct=(1,1))
            ax1.plot(self.times*1e9, SNRn, "-", color='blue', label="Naïve SNR")
            ax2.plot(self.times*1e9, dSNRn, "--", color='blue', label=r"Naïve  $\frac{d SNR}{dt}$")
        
        ax1.legend()
        ax2.legend()
        #plt.legend()
        
    def get_fidelity_time(self, er=None, ef=None, nu=1, weightFct='optimal'):
        dSNR, SNR = self.get_SNR_time(er=er, ef=ef, nu=nu, weightFct=weightFct)
        return (1-(erfc(SNR/2)/2))
        
    def plot_fidelity_time(self,er=None, ef=None, nu=1, naiive=False):
        fid = self.get_fidelity_time(er=er, ef=ef, nu=nu)
        #fig, ax1 = plt.subplots()
        plt.plot(self.times*1e9, fid, "-", color='red', label="Optimal SNR")
        plt.xlabel('Time [ns]')
        plt.ylabel('Fidelity')
        
        if naiive:    
            fidn = self.get_fidelity_time(er=er, ef=ef, nu=nu, weightFct=(1,1))
            plt.plot(self.times*1e9, fidn, "-", color='blue', label="Naïve SNR")
            
        plt.legend()
            
    def optimized_snr_slope(self, eta=1.0):
        "Obtain slope a of SNR using optimized weight functions"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(er=1)
        return 4*self.params.kappa_r*eta*np.sum(np.absolute(r0 - r1)**2)*self.params.dt
    
    def naive_weights_snr(self, eta=1.0, integration_factor=1):
        "obtain slope a of SNR using constant weight function (with optimal phase):"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(er=1)
        sig = 2*np.sqrt(eta*self.params.kappa_r) * np.absolute(np.sum((r0 - r1)))*self.params.dt
        noise = np.sqrt(self.params.tmax * integration_factor)
        return (sig/noise)**2
        
    def dephasing_strength(self):
        "Obtain dephasing rate b as"
        "contrast = exp(-b * e0**2)"        
        r0, r1 = self.get_intercavity_fields()
        return 2*self.chi*np.sum(np.imag(r1 * r0.conj())*self.params.dt)
    
    def estimate_eta(self, real_eta=1.0,  weight_method='optimal', integration_factor=1):
        if weight_method == 'optimal':
            snr_a = self.optimized_snr_slope(real_eta)
        else:
            snr_a = self.naive_weights_snr(real_eta, integration_factor)
        deph = self.dephasing_strength()
        return snr_a/(8*deph)
    

#####################################################################################
def avPhNum(f):
    n = np.average(np.abs(f)**2)
    return n

def isArray(a):
    return hasattr(a, '__len__')
#####################################################################################


def vectorfields_filterDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func]
    """
    alpha, beta = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, ef_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(ef_func):
        ef = ef_func(t)
    elif isArray(ef_func):    
        l=len(ef_func)
        if  (l*1.01) > i:
            ef = ef_func[min(i,l-1)]
        else:
            ef = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        ef = ef_func
        
    #print(i, ef)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta
    db = - 1j*(w_f-w_d+chi)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta - 1j*ef
    
    return [da, db]


def vectorfields_resonatorDrive(t, y, par):
    """
    Defines the differential equations for the filter-resonator system.

    Arguments:
        y :  vector of the state variables:
                  y = [alpha, beta, i]
        t :  time
        par :  vector of the parameters:
                  p = [w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func]
    """
    alpha, beta  = y
    w_d, w_r, w_f, chi, G, kappa_f, kappa_r, dt, er_func = par
    
    i=max(int(np.round(t/dt))-1, 0)
    if callable(er_func):
        er = er_func(t)
    elif isArray(er_func):
        l=len(er_func)
        if  (l*1.01) > i:
            er = er_func[min(i,l-1)]
        else:
            er = 0
            warnings.warn('ERROR: Not enough drive data! %d/%d'%(i, l))
    else:
        er = er_func
    
    #print(i, er)
    
    da = - 1j*(w_r-w_d+chi)*alpha - 1j*G*beta - 1j*er
    db = - 1j*(w_f-w_d+chi)*beta - 1j*np.conj(G)*alpha - (kappa_f/2)*beta
    
    return [da, db]

def solve_ODE(qubit, diffeq):
    resp_alpha = np.zeros_like(qubit.times, dtype=complex)
    resp_beta = np.zeros_like(qubit.times, dtype=complex)
    t=0
    i=0
    while (diffeq.successful() and ((diffeq.t+qubit.params.dt) < qubit.params.tmax)):
        i += 1
        t = qubit.times[i]
        diffeq.integrate(t)
        a = complex(diffeq.y[0])
        b = complex(diffeq.y[1])
        #print(str(i)+"/"+str(len(qubit.times)), t, a, b)
        resp_alpha[i] = a
        resp_beta[i] = b
    
    return (resp_alpha, resp_beta)

def solve_ODE2(qubit, diffeq):
    pass

def solve_ODE_self(qubit, func, par, drive):
    resp_alpha = np.zeros_like(qubit.times, dtype=complex)
    resp_beta = np.zeros_like(qubit.times, dtype=complex)
    t=0
    
    a=0
    b=0
    par.append(0)
    dt = qubit.params.dt
    dal=dt
    dbl=dt
    for i, t in enumerate(qubit.times):
        par[-1] = drive[i]
        
        da, db = func(t=t, y=(a,b), par=par)
        #if max(abs(da)/abs(dal), abs(db)/abs(dbl)) > 10:
        #    da2, db2 = func(t=(t-dt/2), y=(a,b), par=par)
        ##    a += da*dt/2
        #    b += db*dt/2
        #    da2, db2 = func(t=t, y=(a,b), par=par)
        #    a += da*dt/2
        #    b += db*dt/2
        
        resp_alpha[i] = a
        resp_beta[i] = b
        a += complex(da)*dt
        b += complex(db)*dt
        #dal, dbl = (da, db)
    
    return (resp_alpha, resp_beta)
