#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np
from numpy import pi
import matplotlib.pyplot as plt
from IPython.core.display import display, HTML



#####################################################################################
class PurcellSetupParameterSet:
    w_rE = None
    w_rG = None
    chi = None
    w_r = None
    w_q = None
    w_f = None
    Qf = None
    kappa_r = None
    kappa_f = None
    G = None
    Gstar = None #Complex conjugate of G
    kappa_q = None
    w_d = None
    delta_rd = None
    delta_fd = None
    delta_wr = None
    delta = None
    tmax = None
    t0 = None
    t1 = None
    dt = None
    t_factor = None
    
    def _comp(self, a, b):
        return (abs(a-b) < abs(a*0.01))
    
    def is_well_defined(self):
        if not(self.delta_fd is None or self.kappa_f is None):
            self.t_factor = 1/(self.kappa_f/2+1j*self.delta_fd)
            self.Gstar = np.conjugate(self.G)
            
            notNone = not(self.w_rE is None or
                   self.w_rG is None or
                   self.chi is None or
                   self.w_r is None or
                   self.w_q is None or
                   self.w_f is None or
                   self.Qf is None or
                   self.kappa_r is None or
                   self.G is None or
                   self.kappa_q is None or
                   self.w_d is None or
                   self.delta_rd is None or
                   self.delta_wr is None or
                   self.delta is None or
                   self.tmax is None or
                   self.t0 is None or
                   self.t1 is None or
                   self.dt is None)
            
            if notNone:
                dt = min(1/self.kappa_r, self.tmax, self.t0, self.t1)/10
                sanity1 = ((self.dt <= dt) and (self.t0 <= self.tmax) and (self.t1 <= self.tmax) and (self.t0 <= self.t1))
                if not sanity1: print('san1')
                
                sanity2 = ( self._comp((self.w_r + self.chi), self.w_rG) and self._comp((self.w_r - self.chi), self.w_rE) )
                if not sanity2: print('san2')
                
                sanity3 = (self._comp((self.w_f/self.kappa_f), self.Qf) and self._comp(self.G, np.sqrt((self.kappa_f*self.kappa_r)*(1+(2*(self.w_r-self.w_f)/self.kappa_f)**2)/4)))
                if not sanity3: print('san3')
                
                sanity4 = (self._comp(self.delta_rd, (self.w_r - self.w_d)) and self._comp(self.delta_fd, (self.w_f - self.w_d)) and self._comp(self.delta_wr, (-(self.delta_fd/self.kappa_f) * self.kappa_r)))
                if not sanity3: print('san4')
                
                sanity5 = self._comp(self.delta, (self.delta_rd + self.delta_wr))
                if not sanity5: print('san5')
                
                return (sanity1 and sanity2 and sanity3 and sanity4 and sanity5)
        return False
    
    def printParams(self):
        #Print all the parameters in a small table
        if self.is_well_defined():
            beta_f = -1j * self.t_factor * self.Gstar
            display(HTML("<table>"
                +"<tr><th>Variable</th><th></th><th>Value</th><th>Unit</th></tr>"
                +"<tr><td>(𝜔<sub>r</sub> ± 𝜒)/2𝜋</td><td>=</td><td>(%.5f ± %.5f)</td><td>GHz</td></tr>" % (((self.w_r*1e-9)/(2*pi)), ((self.chi*1e-9)/(2*pi)))
                +"<tr><td>𝜔<sub>f</sub>/2𝜋</td><td>=</td><td>%+.4f</td><td>GHz</td></tr>" % ((self.w_f*1e-9)/(2*pi))
                +"<tr><td>𝜔<sub>q</sub>/2𝜋</td><td>=</td><td>%+.4f</td><td>GHz</td></tr>" % ((self.w_q*1e-9)/(2*pi))
                +"<tr><td>𝜔<sub>d</sub>/2𝜋</td><td>=</td><td>%+.5f</td><td>GHz</td></tr>" % ((self.w_d*1e-9)/(2*pi))
                +"<tr><td>𝛥<sub>rd</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_rd*1e-6)/(2*pi))
                +"<tr><td>𝛥<sub>fd</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_fd*1e-6)/(2*pi))
                +"<tr><td>𝛥/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta*1e-6)/(2*pi))
                +"<tr><td>𝛿𝜔<sub>r</sub>/2𝜋</td><td>=</td><td>%+.3f</td><td>MHz</td></tr>" % ((self.delta_wr*1e-6)/(2*pi))
                +"<tr><td>¹/𝜅<sub>f</sub></td><td>=</td><td>%+.4f</td><td>ns</td></tr>" % (1e9/self.kappa_f)
                +"<tr><td>¹/𝜅<sub>r</sub></td><td>=</td><td>%+.4f</td><td>ns</td></tr>" % (1e9/self.kappa_r)
                +"<tr><td>T<sub>1</sub> = 1/𝜅<sub>q</sub></td><td>=</td><td>%+.4f</td><td>µs</td></tr>" % (1e6/self.kappa_q)
                +"<tr><td>|𝜒|/𝜅<sub>r</sub></td><td>=</td><td>%+.4f</td><td></td></tr>" % (abs(self.chi)/self.kappa_r)
                +"<tr><td>G/2𝜋</td><td>=</td><td>%+.2f</td><td>MHz</td></tr>" % (self.G/(2*pi*1e6))
                +"<tr><td>Q<sub>f</sub></td><td>=</td><td>%+.2f</td><td></td></tr>" % self.Qf
                +"<tr><td>𝛽 = a ∙ 𝛼 ; a</td><td>=</td><td>(%+f, %+f j)</td><td></td></tr>" % (beta_f.real, beta_f.imag)
                +"<tr><td>T</td><td>=</td><td>%+.2f</td><td>ns</td></tr>" % (self.tmax*1e9)
                +"<tr><td>dT</td><td>=</td><td>%+.4f</td><td>ns</td></tr>" % (self.dt*1e9)
                +"</table>"))
        else:
            print('Not fully defined yet.')



#####################################################################################
class SquareDepletionMeasurementPulsePurcell:
    def __init__(self, params, depletion=True):
        if not params.is_well_defined():
            print('Error: Given parameters are not well-defined!')
        
        self.params = params
        self.times = np.arange(0, self.params.tmax, self.params.dt)
        
        self.ntimes = len(self.times)
        
        #Green's function of the resonator
        self.greens_function_zero = -1j*np.exp((-1j*(params.delta + params.chi) - params.kappa_r/2)*self.times)
        self.greens_function_one = -1j*np.exp((-1j*(params.delta - params.chi) - params.kappa_r/2)*self.times)

        #components of measurement pulse
        
        self.window0 = np.zeros(self.ntimes)
        self.window0[np.where(self.times < params.t0)] = 1
        
        
        "depletion pulse"
        if depletion:
            self.window1 = np.zeros(self.ntimes)
            self.window1[np.where((self.times >= params.t0) & (self.times < params.t1))] = 1
            self.window2 = np.zeros(self.ntimes)
            self.window2[np.where(params.t1 <= self.times)] = 1
            #solve linear equation to find depletion parameters

            A = np.array([
                [np.dot(self.window1, self.greens_function_zero[::-1]),
                 np.dot(self.window2, self.greens_function_zero[::-1])], 
                [np.dot(self.window1, self.greens_function_one[::-1]), 
                 np.dot(self.window2, self.greens_function_one[::-1])],     
            ])

            b = np.array([
                np.dot(self.window0, self.greens_function_zero[::-1]),
                np.dot(self.window0, self.greens_function_one[::-1]), 
            ])
	
            "depletion parameters"
            self.e1, self.e2 = np.linalg.solve(A, -b)
            self.edrive = self.window0 + self.e1*self.window1 + self.e2*self.window2
        else:
            self.edrive = self.window0

        "inter-cavity field for 0 and 1 state"
        self.resp0 = self.params.dt*np.convolve(self.greens_function_zero, self.edrive)[:self.ntimes]
        self.resp1 = self.params.dt*np.convolve(self.greens_function_one, self.edrive)[:self.ntimes]

    def get_intercavity_fields(self, er=None, ef=None):
        er =  self.get_intercavity_amplitude(ef=ef, er=er)
        return er*self.resp0, er*self.resp1
    
    def get_intercavity_amplitude(self, ef=None, er=None):
        if er is None:
            er = -1j*ef*self.params.t_factor*self.params.G
        return er
    
    def get_filter_fields(self,  er=None, ef=None, filterDrive=False):
        er =  self.get_intercavity_amplitude(ef=ef, er=er)
        a0, a1 = self.get_intercavity_fields(er = er)
        
        b0 = np.array(-1j*self.params.t_factor * self.params.Gstar * a0)
        b1 = np.array(-1j*self.params.t_factor * self.params.Gstar * a1)
        
        if filterDrive:
            b0 += er/0.15e9*self.edrive #TODO: WTF factor???
            b1 += er/0.15e9*self.edrive
        
        return b0, b1
    
    
    def plot_alphas_time(self, er=None, ef=None, plot_imag=True):
        er =  self.get_intercavity_amplitude(ef=ef, er=er)
        r0, r1 = self.get_intercavity_fields(er)
        
        plt.plot(self.times*1e9, r0.real, "-", color='red', label=r"Re($\alpha_0$)")
        plt.plot(self.times*1e9, r1.real, "-", color='green', label=r"Re($\alpha_1$)")
        if plot_imag:
            plt.plot(self.times*1e9, r0.imag, "--", color='red', label=r"Im($\alpha_0$)")
            plt.plot(self.times*1e9, r1.imag, "--", color='green', label=r"Im($\alpha_1$)")
        
        plt.xlabel("Time [ns]")
        plt.legend()

    def plot_alphas_complex(self, er=None, ef=None):
        a0, a1 = self.get_intercavity_fields(er=er, ef=ef)

        x0 = np.real(a0)
        y0 = np.imag(a0)
        x1 = np.real(a1)
        y1 = np.imag(a1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.1
        xd = max(abs(xmin), abs(xmax))*0.1
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\alpha_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\alpha_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()


    def plot_betas_complex(self, er=None, ef=None, filterDrive=False):
        b0, b1 = self.get_filter_fields(er, ef, filterDrive=filterDrive)
        
        x0 = np.real(b0)
        y0 = np.imag(b0)
        x1 = np.real(b1)
        y1 = np.imag(b1)
        xmin = min(np.min(x0), np.min(x1))
        xmax = max(np.max(x0), np.max(x1))
        ymin = min(np.min(y0), np.min(y1))
        ymax = max(np.max(y0), np.max(y1))
        yd = max(abs(ymin), abs(ymax))*0.4
        xd = max(abs(xmin), abs(xmax))*0.4
	
        plt.plot([xmin-xd, xmax+xd], [0,0], color='darkgrey')
        plt.plot([0,0], [ymin-yd, ymax+yd], color='darkgrey')
        
        plt.plot(x0, y0, "-", color='red', label=r"$\beta_0$")
        plt.plot(x1, y1, "-", color='green', label=r"$\beta_1$")
        
        plt.xlabel(s="Real")
        plt.ylabel(s="Imaginary")
        plt.legend()
    

    def plot_optimal_weight_function(self, e0=1, iq=False):
        r0, r1 = self.get_intercavity_fields()

        weight_function = r1 - r0
        self.weight_function_I = weight_function.real
        self.weight_function_Q = weight_function.imag
            
        if iq:
            plt.plot(weight_function.real, weight_function.imag)
            plt.xlabel(r"Re($\alpha_e - \alpha_g$)")
            plt.ylabel(r"Im($\alpha_e - \alpha_g$)")
        else:
            plt.plot(self.times, weight_function.real, label=r"Re($\alpha_e - \alpha_g$)")
            plt.plot(self.times, weight_function.imag, label=r"Im($\alpha_e - \alpha_g$)")
            plt.xlabel("Time [ns]")
            plt.legend()
            
    def optimized_snr_slope(self, eta=1.0):
        "Obtain slope a of SNR using optimized weight functions"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(e0=1)
        return 4*self.params.kappa_r*eta*np.sum(np.absolute(r0 - r1)**2)*self.params.dt
    
    def naive_weights_snr(self, eta=1.0, integration_factor=1):
        "obtain slope a of SNR using constant weight function (with optimal phase):"
        "SNR² = a*e0"
        r0, r1 = self.get_intercavity_fields(e0=1)
        sig = 2*np.sqrt(eta*self.params.kappa_r) * np.absolute(np.sum((r0 - r1)))*self.params.dt
        noise = np.sqrt(self.params.tmax * integration_factor)
        return (sig/noise)**2
        
    def dephasing_strength(self):
        "Obtain dephasing rate b as"
        "contrast = exp(-b * e0**2)"        
        r0, r1 = self.get_intercavity_fields()
        return 2*self.chi*np.sum(np.imag(r1 * r0.conj())*self.params.dt)
    
    def estimate_eta(self, real_eta=1.0,  weight_method='optimal', integration_factor=1):
        if weight_method == 'optimal':
            snr_a = self.optimized_snr_slope(real_eta)
        else:
            snr_a = self.naive_weights_snr(real_eta, integration_factor)
        deph = self.dephasing_strength()
        return snr_a/(8*deph)